try:
    import os, sys, json
    import numpy as np
    import math
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import onnx
        from onnx_tf.backend import prepare
        from . import tf_util
    HAS_TF_ONNX = True

except Exception as e:
    HAS_TF_ONNX = False


def convert_bounding_boxes_to_coord_list(bounding_boxes):
    """
    convert bounding box numpy array to python list of point arrays
    :param bounding_boxes: numpy array of shape [n, 4]
    :return: python array of point numpy arrays, each point array is in shape [4,2]
    """
    num_bounding_boxes = bounding_boxes.shape[0]
    bounding_box_coord_list = []
    for i in range(num_bounding_boxes):
        coord_array = np.empty(shape=(4, 2), dtype=float)
        coord_array[0][0] = bounding_boxes[i][0]
        coord_array[0][1] = bounding_boxes[i][1]

        coord_array[1][0] = bounding_boxes[i][0]
        coord_array[1][1] = bounding_boxes[i][3]

        coord_array[2][0] = bounding_boxes[i][2]
        coord_array[2][1] = bounding_boxes[i][3]

        coord_array[3][0] = bounding_boxes[i][2]
        coord_array[3][1] = bounding_boxes[i][1]

        bounding_box_coord_list.append(coord_array)

    return bounding_box_coord_list


def _create_anchors(anc_grids, anc_zooms, anc_ratios):
    anchor_scales = [(anz * i, anz * j) for anz in anc_zooms for (i, j) in anc_ratios]

    _anchors_per_cell = len(anchor_scales)

    anc_offsets = [1 / (o * 2) for o in anc_grids]

    anc_x = np.concatenate(
        [
            np.repeat(np.linspace(ao, 1 - ao, ag), ag)
            for ao, ag in zip(anc_offsets, anc_grids)
        ]
    )
    anc_y = np.concatenate(
        [
            np.tile(np.linspace(ao, 1 - ao, ag), ag)
            for ao, ag in zip(anc_offsets, anc_grids)
        ]
    )
    anc_ctrs = np.repeat(np.stack([anc_x, anc_y], axis=1), _anchors_per_cell, axis=0)

    anc_sizes = np.concatenate(
        [
            np.array(
                [[o / ag, p / ag] for i in range(ag * ag) for o, p in anchor_scales]
            )
            for ag in anc_grids
        ]
    )

    _grid_sizes = np.expand_dims(
        np.concatenate(
            [
                np.array([1 / ag for i in range(ag * ag) for o, p in anchor_scales])
                for ag in anc_grids
            ]
        ),
        1,
    )

    _anchors = np.concatenate([anc_ctrs, anc_sizes], axis=1)

    return _anchors, _grid_sizes


def calculate_rectangle_size_from_batch_size(batch_size):
    """
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    """
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    """
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    """
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width


def tile_to_batch(
    pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs
):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs["batch_height"]
        batch_width = kwargs["batch_width"]
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(
        shape=(batch_width * batch_height, band_count, model_height, model_width),
        dtype=pixel_type,
    )
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[
            :,
            y * inner_height : y * inner_height + model_height,
            x * inner_width : x * inner_width + model_width,
        ]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[
            b, :, : sub_pixel_block_shape[1], : sub_pixel_block_shape[2]
        ] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(
        shape=(bands, inner_height * batch_height, inner_width * batch_width),
        dtype=batch.dtype,
    )

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[
            :,
            y * inner_height : (y + 1) * inner_height,
            x * inner_width : (x + 1) * inner_width,
        ] = batch[b]

    return tile


def remove_bounding_boxes_in_padding(
    bounding_boxes,
    scores,
    classes,
    image_height,
    image_width,
    padding,
    batch_height=1,
    batch_width=1,
):
    """

    :param bounding_boxes: the batch of bounding boxes, shape=[B,N,4]
    :param scores: the batch of box scores, shape=[B,N]
    :param classes: the batch of labels, shape=[B,N]
    :param image_height: model height
    :param image_width: model width
    :param padding:
    :param batch_height:
    :param batch_width:
    :return:
    """
    keep_indices = np.where(
        (bounding_boxes[:, :, 0] < image_height - padding)
        & (bounding_boxes[:, :, 1] < image_width - padding)
        & (bounding_boxes[:, :, 2] > padding)
        & (bounding_boxes[:, :, 3] > padding)
    )

    inner_width = image_width - 2 * padding
    inner_height = image_height - 2 * padding

    # convert coordinates in the batch to super tile and then filter by the keep_indices
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        bounding_boxes[b, :, [0, 2]] = bounding_boxes[b, :, [0, 2]] + y * inner_height
        bounding_boxes[b, :, [1, 3]] = bounding_boxes[b, :, [1, 3]] + x * inner_width

    bounding_boxes = bounding_boxes[keep_indices]
    scores = scores[keep_indices]
    classes = classes[keep_indices]

    return bounding_boxes, scores, classes


class ChildObjectDetector:
    def initialize(self, model, model_as_file):
        if not HAS_TF_ONNX:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        from arcgis.learn.models import SingleShotDetector

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.loads(model)

        model_path = self.json_info["ModelFile"]
        anc_grids = self.json_info["Grids"]
        anc_zooms = self.json_info["Zooms"]
        anc_ratios = self.json_info["Ratios"]
        self.batch_size = self.json_info["BatchSize"]

        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        model_onnx = onnx.load(model_path)
        self.tf_rep_graph = prepare(model_onnx)

        self._anchors, self._grid_sizes = _create_anchors(
            anc_grids, anc_zooms, anc_ratios
        )

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": self.json_info["ImageHeight"] // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "threshold",
                    "dataType": "numeric",
                    "value": 0.5,
                    "required": False,
                    "displayName": "Confidence Score Threshold [0.0, 1.0]",
                    "description": "Confidence score threshold value [0.0, 1.0]",
                },
                {
                    "name": "nms_overlap",
                    "dataType": "numeric",
                    "value": 0.1,
                    "required": False,
                    "displayName": "NMS Overlap",
                    "description": "Maximum allowed overlap within each chip",
                },
                {
                    "name": "exclude_pad_detections",
                    "dataType": "string",
                    "required": False,
                    "domain": ("True", "False"),
                    "value": "True",
                    "displayName": "Filter Outer Padding Detections",
                    "description": "Filter detections which are outside the specified padding",
                },
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.padding = int(
            scalars.get("padding", self.json_info["ImageHeight"] // 4)
        )  ## Default padding Imageheight//4.
        self.nms_overlap = float(
            scalars.get("nms_overlap", 0.1)
        )  ## Default 0.1 NMS Overlap.
        self.thres = float(scalars.get("threshold", 0.5))  ## Default 0.5 threshold.
        self.filter_outer_padding_detections = scalars.get(
            "exclude_pad_detections", "True"
        ).lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  ## Default value True

        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        return {
            "extractBands": tuple(self.json_info["ExtractBands"]),
            "padding": self.padding,
            "threshold": self.thres,
            "nms_overlap": self.nms_overlap,
            "tx": tx,
            "ty": ty,
            "fixedTileSize": 1,
        }

    def vectorize(self, **pixelBlocks):  # 8 x 3 x 224 x 224
        input_image = pixelBlocks["raster_pixels"]
        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        bounding_boxes, scores, classes = tf_util.detect_objects_image_space(
            self.tf_rep_graph,
            batch,
            self._anchors,
            self._grid_sizes,
            classes=[clas["Name"] for clas in self.json_info["Classes"]],
            nms_overlap=self.nms_overlap,
            thres=self.thres,
        )

        return convert_bounding_boxes_to_coord_list(bounding_boxes), scores, classes



from dataclasses import dataclass, asdict, field
from typing import Dict, Union, Optional, Any, ClassVar


@dataclass
class _HttpAuthenticationType:
    _auth_type: ClassVar[str]

    def _build(self, feed_or_source_name: str) -> Dict[str, str]:
        raise NotImplementedError


@dataclass
class NoAuth(_HttpAuthenticationType):
    """This dataclass is used to specify the no HTTP authentication scenario."""

    _auth_type: ClassVar[str] = "noauth"

    def _build(self, feed_or_source_name: str) -> Dict[str, str]:
        return {f"{feed_or_source_name}.httpAuthenticationType": self._auth_type}


@dataclass
class BasicAuth(_HttpAuthenticationType):
    """
    This dataclass is used to specify a Basic HTTP Authentication scenario using username and password.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    username               String. Username for basic authentication.
    ------------------     --------------------------------------------------------------------
    password               String. Password for basic authentication.
    ==================     ====================================================================
    """

    _auth_type: ClassVar[str] = "basicauth"

    username: str
    password: str

    def _build(self, feed_or_source_name: str) -> Dict[str, str]:
        return {
            f"{feed_or_source_name}.httpAuthenticationType": self._auth_type,
            f"{feed_or_source_name}.username": self.username,
            f"{feed_or_source_name}.password": self.password,
        }


@dataclass
class CertificateAuth(_HttpAuthenticationType):
    """
    This dataclass is used to specify a Basic HTTP Authentication scenario using username and password.

    =======================     ====================================================================
    **Parameter**                **Description**
    -----------------------     --------------------------------------------------------------------
    pfx_file_http_location      String. HTTP path of the PFX file.
    -----------------------     --------------------------------------------------------------------
    password                    String. Password for certificate authentication.
    =======================     ====================================================================
    """

    _auth_type: ClassVar[str] = "certificateauth"

    pfx_file_http_location: str
    password: str

    def _build(self, feed_or_source_name: str) -> Dict[str, str]:
        return {
            f"{feed_or_source_name}.httpAuthenticationType": self._auth_type,
            f"{feed_or_source_name}.password": self.password,
            f"{feed_or_source_name}.pfxLocation": self.pfx_file_http_location,
        }



from pathlib import Path
import types
import os
import torch
import torch.nn.functional as F
import numpy as np
import mmcv
from mmdet3d.models import build_detector
import logging
from mmcv.runner import auto_fp16


def get_backbone_channel(model_cfg, data):
    temp_model = build_detector(model_cfg).to(data.device)

    # x_batch, _ = data.one_batch(detach=False)
    x_batch = torch.rand(
        (10000, data.num_features), dtype=torch.float32, device=data.device
    )
    x_batch[:, :3] = (x_batch[:, :3] - 0.5) / 0.5
    x_batch[:, :3] *= data.scale_factor

    voxels, num_points, coors = temp_model.voxelize([x_batch])
    voxel_features = temp_model.voxel_encoder(voxels, num_points, coors)
    batch_size = coors[-1, 0].item() + 1
    backbone_feature = temp_model.middle_encoder(voxel_features, coors, batch_size)

    return backbone_feature.shape[1]


def get_voxel_size(voxel_parms, data):
    pc_range = np.array(data.range)
    pc_lwh = pc_range[3:] - pc_range[:3]
    # keep the minimum resolution of point cloud grid to (200, 200) in x, y direction
    tile_voxel_size = ([0.005, 0.005, 0.02] * pc_lwh).tolist()
    default_voxel_size = list(map(min, zip(tile_voxel_size, data.voxel_size)))
    voxel_size = voxel_parms.get("voxel_size", None)
    if voxel_size is None:
        voxel_size = default_voxel_size
    else:
        if len(voxel_size) != 3:
            raise Exception("voxel_size list should contain 3 items.")
        # check if given voxel_size is creating at least 64x64x8 voxels
        grid_size = pc_lwh / voxel_size
        if min(grid_size) < 64:
            raise Exception(f"The size {voxel_size} of the voxel is too big.")

    grid_size = torch.tensor(pc_lwh / voxel_size).round().long().tolist()[::-1]

    return voxel_size, grid_size


def get_max_voxels(voxel_parms, grid_size):
    no_of_voxels = np.prod(grid_size, dtype=np.uint64).tolist()
    max_voxels = voxel_parms.get("max_voxels", None)
    if max_voxels is None:
        max_voxels = list(
            map(max, [20000, 40000], [no_of_voxels // 3000, no_of_voxels // 2000])
        )
    else:
        if len(max_voxels) != 2:
            raise Exception("max_voxels list should contain 2 items.")
        if min(max_voxels) < 20000:
            raise Exception("The minimum values in max_voxel should be at least 20000.")
    return max_voxels


def set_voxel_info(voxel_parms, data):
    voxel_size, grid_size = get_voxel_size(voxel_parms, data)
    max_voxels = get_max_voxels(voxel_parms, grid_size)
    voxel_points = voxel_parms.get("voxel_points", None)
    if voxel_points is None:
        voxel_points = max(10, int(data.no_of_points_per_tile // (max_voxels[0] * 0.3)))
    if voxel_points < 10:
        raise Exception("voxel_points should be greater than or equal to 10.")

    voxel_parms["voxel_size"] = voxel_size
    voxel_parms["sparse_shape"] = grid_size
    voxel_parms["max_voxels"] = max_voxels
    voxel_parms["voxel_points"] = voxel_points

    return voxel_parms


def model_config(model_cfg, data, **kwargs):
    voxel_parms = kwargs.get("voxel_parms", {})
    voxel_parms = set_voxel_info(voxel_parms, data)

    model_cfg.voxel_layer.voxel_size = voxel_parms["voxel_size"]
    model_cfg.voxel_layer.max_voxels = voxel_parms["max_voxels"]
    model_cfg.voxel_layer.max_num_points = voxel_parms["voxel_points"]
    model_cfg.voxel_layer.point_cloud_range = data.range

    model_cfg.voxel_encoder.num_features = data.num_features
    model_cfg.middle_encoder.in_channels = data.num_features

    # set correctly otherwise RuntimeError: CUDA error: an illegal memory access was encountered
    model_cfg.middle_encoder.sparse_shape = voxel_parms["sparse_shape"]

    model_cfg.backbone.in_channels = get_backbone_channel(model_cfg, data)

    model_cfg.bbox_head.num_classes = data.c
    model_cfg.bbox_head.bbox_coder.code_size = 7
    model_cfg.bbox_head.anchor_generator.ranges = data.anchor_range
    model_cfg.bbox_head.anchor_generator.sizes = data.average_box_size
    model_cfg.bbox_head.anchor_generator.rotations = [0.0, 1.57]

    return model_cfg


@auto_fp16(apply_to=("points",))
def forward_modified(self, input):
    if not self.prediction:
        losses = self.forward_train(**input)
        loss, log_vars = self._parse_losses(losses)

        loss = dict(loss=loss, log_vars=log_vars)
        output = None
        if not self.training:
            output = self.simple_test(input["points"], input["img_metas"])
        return output, loss
    else:
        return self.simple_test(input["points"], input["img_metas"])


@auto_fp16()
def forward_neck(self, x):
    assert len(x) == len(self.in_channels)
    ups = [deblock(x[i]) for i, deblock in enumerate(self.deblocks)]
    size = ups[0].shape[-2:]
    ups = [F.interpolate(up, size, mode="bilinear", align_corners=False) for up in ups]
    if len(ups) > 1:
        out = torch.cat(ups, dim=1)
    else:
        out = ups[0]
    return [out]


def get_model(data, **kwargs):
    logging.disable(logging.WARNING)

    config = kwargs.get("model")
    checkpoint = kwargs.get("model_weight", False)

    if os.path.exists(Path(config)):
        cfg = mmcv.Config.fromfile(config)
    else:
        import arcgis

        cfg_abs_path = (
            Path(arcgis.__file__).parent
            / "learn"
            / "_mmdet3d_config"
            / (config + ".{}".format("py"))
        )
        cfg = mmcv.Config.fromfile(cfg_abs_path)
        checkpoint = cfg.get("checkpoint", False)

    cfg.model = model_config(cfg.model, data, **kwargs)

    model = build_detector(cfg.model)

    if checkpoint:
        mmcv.runner.load_checkpoint(
            model, checkpoint, "cpu", False, logging.getLogger()
        )

    model.forward = types.MethodType(forward_modified, model)
    if cfg.model.neck.type == "SECONDFPN":
        model.neck.forward = types.MethodType(forward_neck, model.neck)

    model.prediction = False

    logging.disable(0)

    return model, cfg



import os
import uuid
import copy
import shutil
import tempfile
import logging
from arcgis._impl.common._clone import CloneNode, _deep_get
from arcgis._impl.common._clone import (
    _search_org_for_existing_item,
    _share_item_with_groups,
)

try:
    import ujson as json
except ImportError:
    import json


class _StoryMapDefinition(CloneNode):
    """
    Represents the definition of a quick capture project within ArcGIS Online or Portal.
    """

    def __init__(
        self,
        target,
        info,
        data,
        sharing=None,
        thumbnail=None,
        portal_item=None,
        folder=None,
        item_extent=None,
        search_existing=True,
        owner=None,
        **kwargs,
    ):
        clone_mapping = kwargs.get("clone_mapping")
        super().__init__(target, clone_mapping, search_existing)
        self._preserve_item_id = kwargs.pop("preserve_item_id", False)
        self.info = info
        self._data = data
        self.resolved = False
        self.sharing = sharing
        if not self.sharing:
            self.sharing = {"access": "private", "groups": []}
        self.thumbnail = thumbnail
        self._item_property_names = [
            "title",
            "type",
            "description",
            "snippet",
            "tags",
            "culture",
            "accessInformation",
            "licenseInfo",
            "typeKeywords",
            "extent",
            "url",
            "properties",
        ]
        self.portal_item = portal_item
        self.folder = folder
        self.owner = owner
        self.item_extent = item_extent
        self.created_items = []
        self.resources = kwargs.pop("resources", None)
        self._clone_mapping = clone_mapping
        self.clone_mapping = self._clone_mapping

    @property
    def data(self):
        """Gets the data of the item"""
        return copy.deepcopy(self._data)

    def _add_new_item(self, item_properties, data=None):
        """Add the new item to the portal"""
        thumbnail = self.thumbnail
        if not thumbnail and self.portal_item:
            temp_dir = os.path.join(self._temp_dir.name, self.info["id"])
            if not os.path.exists(temp_dir):
                os.makedirs(temp_dir)
            thumbnail = self.portal_item.download_thumbnail(temp_dir)
        item_id = None
        if self._preserve_item_id and self.target._portal.is_arcgisonline:
            item_id = self.portal_item.itemid
        new_item = self.target.content.add(
            item_properties=item_properties,
            data=data,
            thumbnail=thumbnail,
            folder=self.folder,
            owner=self.owner,
            item_id=item_id,
        )
        if self.portal_item.url:
            if self.target._portal.is_arcgisonline:
                url = f"https://storymaps.arcgis.com/stories/{new_item.id}"
            else:
                url = f"{self.target._portal.url}/apps/storymaps/stories/{new_item.id}"
            new_item.update({"url": url})
        self.created_items.append(new_item)
        self._clone_resources(new_item)
        return new_item

    def _clone_resources(self, new_item):
        """Add the resources to the new item"""
        if self.portal_item:
            resources = self.portal_item.resources
            resource_list = resources.list()
            if len(resource_list) > 0:
                resources_dir = os.path.join(
                    self._temp_dir.name, self.info["id"], "resources"
                )
                if not os.path.exists(resources_dir):
                    os.makedirs(resources_dir)
                for resource in resource_list:
                    resource_name = resource["resource"]
                    folder_name = None
                    if len(resource_name.split("/")) == 2:
                        folder_name, resource_name = resource_name.split("/")
                    elif len(resource_name.split("/")) > 2:
                        folder_name = os.path.dirname(resource_name)
                        resource_name = os.path.basename(resource_name)
                    resource_path = resources.get(
                        resource["resource"], False, resources_dir, resource_name
                    )
                    new_item.resources.add(resource_path, folder_name, resource_name)

    def _get_item_properties(self, item_extent=None):
        """Get a dictionary of item properties used in create and update operations."""

        item_properties = {}
        for property_name in self._item_property_names:
            if property_name in self.info and self.info[property_name] is not None:
                item_properties[property_name] = self.info[property_name]

        type_keywords = item_properties["typeKeywords"]
        for keyword in list(type_keywords):
            if keyword.startswith("source-"):
                type_keywords.remove(keyword)

        tags = item_properties["tags"]
        type_keywords.append("source-{0}".format(self.info["id"]))
        item_properties["typeKeywords"] = ",".join(item_properties["typeKeywords"])
        item_properties["tags"] = ",".join(tags)

        item_properties.pop("url", None)
        extent = _deep_get(item_properties, "extent")
        if item_extent is not None and extent is not None and len(extent) > 0:
            item_properties["extent"] = "{0}, {1}, {2}, {3}".format(
                item_extent.xmin, item_extent.ymin, item_extent.xmax, item_extent.ymax
            )

        return item_properties

    def _get_item_data(self):
        temp_dir = os.path.join(self._temp_dir.name, self.info["id"])
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)

        data = self.data
        if not data and self.portal_item:
            data = self.portal_item.download(temp_dir)

        # The item's name will default to the name of the data, if it already exists in the folder we need to rename it to something unique
        name = os.path.basename(data)
        item = next(
            (
                item
                for item in self.target.users.get(self.owner).items(folder=self.folder)
                if item["name"] == name
            ),
            None,
        )
        if item:
            new_name = "{0}_{1}{2}".format(
                os.path.splitext(name)[0],
                str(uuid.uuid4()).replace("-", ""),
                os.path.splitext(name)[1],
            )
            new_path = os.path.join(temp_dir, new_name)
            os.rename(data, new_path)
            data = new_path

        return data

    def clone(self):
        """
        Clone the StoryMap v2.0 Item in the target organization.
        """
        new_item = None
        original_item = self.info
        if self._search_existing:
            new_item = _search_org_for_existing_item(self.target, self.portal_item)
        if not new_item:
            # Get the item properties from the original item to be applied when the new item is created
            item_properties = self._get_item_properties(self.item_extent)
            # data = self.data
            resources = self.portal_item.resources
            for res in resources.list():
                if "draft" in res["resource"] and "express" not in res["resource"]:
                    draft_name = res["resource"]

            draft = resources.get(draft_name)

            web_maps = set(
                [
                    v["data"]["itemId"]
                    for k, v in draft["resources"].items()
                    if v["type"].lower().find("webmap") > -1
                ]
            )
            express_maps = set(
                [
                    v["data"]["itemId"]
                    for k, v in draft["resources"].items()
                    if v["type"].lower().find("expressmap") > -1
                ]
            )
            themes = set(
                [
                    v["data"]["themeItemId"]
                    for k, v in draft["resources"].items()
                    if v["type"].lower().find("story-theme") > -1
                    and "themeItemId" in v["data"].keys()
                ]
            )
            webmap_mapper = {}
            for wm in web_maps:
                webmap_to_copy = self.portal_item._gis.content.get(wm)
                if not webmap_to_copy:
                    continue

                # check if webmap is in clone mapping
                if wm in self._clone_mapping["Item IDs"]:
                    new_id = self._clone_mapping["Item IDs"][wm]
                    targ_item = self.target.content.get(new_id)
                    if targ_item:
                        if targ_item.type == webmap_to_copy.type:
                            webmap_mapper[wm] = new_id
                            continue

                # otherwise, clone
                cloned_webmaps = self.target.content.clone_items(
                    [webmap_to_copy],
                    search_existing_items=self._search_existing,
                    folder=self.folder,
                    owner=self.owner,
                    item_extent=self.item_extent,
                    preserve_item_id=self._preserve_item_id,
                )
                if cloned_webmaps:
                    for webmap in cloned_webmaps:
                        self.created_items.append(webmap)
                    webmap_mapper[webmap_to_copy.id] = [
                        i.id
                        for i in cloned_webmaps
                        if (i.type == "Web Map" or i.type == "Web Scene")
                    ]
                    if len(webmap_mapper[webmap_to_copy.id]) == 1:
                        webmap_mapper[webmap_to_copy.id] = webmap_mapper[
                            webmap_to_copy.id
                        ][0]
                # if nothing was cloned, means item exists. grab it
                elif (
                    getattr(webmap_to_copy, "groupDesignations", None) == "livingatlas"
                ):
                    continue
                else:
                    exist_item = _search_org_for_existing_item(
                        self.target, webmap_to_copy
                    )
                    webmap_mapper[wm] = exist_item.id

            if themes:
                for theme in themes:
                    theme_to_copy = self.portal_item._gis.content.get(theme)

                    # check if theme is in clone mapping
                    if theme in self._clone_mapping["Item IDs"]:
                        new_id = self._clone_mapping["Item IDs"][theme]
                        targ_item = self.target.content.get(new_id)
                        if targ_item:
                            if targ_item.type == theme_to_copy.type:
                                webmap_mapper[theme] = new_id
                                continue

                    # otherwise, clone
                    cloned_theme = self.target.content.clone_items(
                        [theme_to_copy], search_existing_items=False
                    )
                    if cloned_theme:
                        for theme in cloned_theme:
                            self.created_items.append(theme)
                    webmap_mapper[theme_to_copy.id] = [
                        i.id for i in cloned_theme if (i.type == "StoryMap Theme")
                    ]
                    if len(webmap_mapper[theme_to_copy.id]) == 1:
                        webmap_mapper[theme_to_copy.id] = webmap_mapper[
                            theme_to_copy.id
                        ][0]
            story_map_text = ""
            if self.data:
                story_map_text = json.dumps(self.data)
                for k, v in webmap_mapper.items():
                    story_map_text = story_map_text.replace(k, v)  # replace the IDs
            if story_map_text:
                item_properties["text"] = story_map_text
            # Add the new item
            new_item = self._add_new_item(item_properties)
            if self.resources:
                new_item.resources.add(self.resources, archive=True)
            for resource in new_item.resources.list():
                if ".json" in resource["resource"]:
                    s_res = json.dumps(
                        new_item.resources.get(resource["resource"]), ensure_ascii=False
                    )
                    for k, v in webmap_mapper.items():
                        s_res = s_res.replace(k, v)
                    res = json.loads(s_res)
                    tfile = tempfile.NamedTemporaryFile(mode="w+", suffix=".json")
                    json.dump(res, tfile)
                    tfile.seek(0)
                    new_item.resources.update(
                        file_name=resource["resource"],
                        file=tfile.name,
                    )
            if new_item.url:
                new_item.update(
                    {"url": new_item.url.replace(self.portal_item.id, new_item.id)}
                )
            with tempfile.NamedTemporaryFile(
                mode="w", suffix=".json", dir=tempfile.gettempdir(), delete=False
            ) as jsonfile:
                jsonfile.write(story_map_text)
                new_item.resources.add(file=jsonfile.name)
                type_keywords = [
                    tk for tk in new_item.typeKeywords if "smdraftresourceid:" not in tk
                ]
                type_keywords.append(
                    f"smdraftresourceid:{os.path.basename(jsonfile.name)}"
                )
                new_item.update({"typeKeywords": type_keywords})
            # express maps
            if len(express_maps) > 0:
                with tempfile.TemporaryDirectory() as d:
                    shutil.unpack_archive(filename=self.resources, extract_dir=d)
                    for expmap in express_maps:
                        express_draft = os.path.join(d, "draft_" + expmap)
                        express_pub = os.path.join(d, "pub_" + expmap)
                        if os.path.isfile(express_pub):
                            shutil.copy(express_pub, express_draft)
                            new_item.resources.add(express_draft)
            _share_item_with_groups(
                new_item, self.sharing, self._clone_mapping["Group IDs"]
            )
        else:
            logging.info(
                self.portal_item.title + " not cloned; already existent in target org."
            )
        self.resolved = True
        self._clone_mapping["Item IDs"][original_item["id"]] = new_item["id"]
        return new_item



"""
MIT License

Copyright (c) 2020 Ziqiang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# based on https://github.com/ZQPei/deep_sort_pytorch
from __future__ import absolute_import
import traceback
import logging

try:
    import torch
    import torchvision.transforms as transforms
    import numpy as np
    import cv2
    import scipy.linalg
    from scipy.optimize import linear_sum_assignment as linear_assignment

    from ._deepsort_train_utils import *

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

INFTY_COST = 1e5


def non_max_suppression(boxes, max_bbox_overlap, scores=None):
    """Suppress overlapping detections.

    Original code from [1]_ has been adapted to include confidence score.

    .. [1] http://www.pyimagesearch.com/2015/02/16/
           faster-non-maximum-suppression-python/

    Examples
    --------

        >>> boxes = [d.roi for d in detections]
        >>> scores = [d.confidence for d in detections]
        >>> indices = non_max_suppression(boxes, max_bbox_overlap, scores)
        >>> detections = [detections[i] for i in indices]

    Parameters
    ----------
    boxes : ndarray
        Array of ROIs (x, y, width, height).
    max_bbox_overlap : float
        ROIs that overlap more than this values are suppressed.
    scores : Optional[array_like]
        Detector confidence score.

    Returns
    -------
    List[int]
        Returns indices of detections that have survived non-maxima suppression.

    """
    if len(boxes) == 0:
        return []

    boxes = boxes.astype(float)
    pick = []

    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2] + boxes[:, 0]
    y2 = boxes[:, 3] + boxes[:, 1]

    area = (x2 - x1 + 1) * (y2 - y1 + 1)
    if scores is not None:
        idxs = np.argsort(scores)
    else:
        idxs = np.argsort(y2)

    while len(idxs) > 0:
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)

        xx1 = np.maximum(x1[i], x1[idxs[:last]])
        yy1 = np.maximum(y1[i], y1[idxs[:last]])
        xx2 = np.minimum(x2[i], x2[idxs[:last]])
        yy2 = np.minimum(y2[i], y2[idxs[:last]])

        w = np.maximum(0, xx2 - xx1 + 1)
        h = np.maximum(0, yy2 - yy1 + 1)

        overlap = (w * h) / area[idxs[:last]]

        idxs = np.delete(
            idxs, np.concatenate(([last], np.where(overlap > max_bbox_overlap)[0]))
        )

    return pick


"""
Table for the 0.95 quantile of the chi-square distribution with N degrees of
freedom (contains values for N=1, ..., 9). Taken from MATLAB/Octave's chi2inv
function and used as Mahalanobis gating threshold.
"""
chi2inv95 = {
    1: 3.8415,
    2: 5.9915,
    3: 7.8147,
    4: 9.4877,
    5: 11.070,
    6: 12.592,
    7: 14.067,
    8: 15.507,
    9: 16.919,
}


class KalmanFilter(object):
    """
    A simple Kalman filter for tracking bounding boxes in image space.

    The 8-dimensional state space

        x, y, a, h, vx, vy, va, vh

    contains the bounding box center position (x, y), aspect ratio a, height h,
    and their respective velocities.

    Object motion follows a constant velocity model. The bounding box location
    (x, y, a, h) is taken as direct observation of the state space (linear
    observation model).

    """

    def __init__(self):
        ndim, dt = 4, 1.0

        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat = np.eye(ndim, 2 * ndim)

        # Motion and observation uncertainty are chosen relative to the current
        # state estimate. These weights control the amount of uncertainty in
        # the model. This is a bit hacky.
        self._std_weight_position = 1.0 / 20
        self._std_weight_velocity = 1.0 / 160

    def initiate(self, measurement):
        """Create track from unassociated measurement.

        Parameters
        ----------
        measurement : ndarray
            Bounding box coordinates (x, y, a, h) with center position (x, y),
            aspect ratio a, and height h.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector (8 dimensional) and covariance matrix (8x8
            dimensional) of the new track. Unobserved velocities are initialized
            to 0 mean.

        """
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        std = [
            2 * self._std_weight_position * measurement[3],
            2 * self._std_weight_position * measurement[3],
            1e-2,
            2 * self._std_weight_position * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            1e-5,
            10 * self._std_weight_velocity * measurement[3],
        ]
        covariance = np.diag(np.square(std))
        return mean, covariance

    def predict(self, mean, covariance):
        """Run Kalman filter prediction step.

        Parameters
        ----------
        mean : ndarray
            The 8 dimensional mean vector of the object state at the previous
            time step.
        covariance : ndarray
            The 8x8 dimensional covariance matrix of the object state at the
            previous time step.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector and covariance matrix of the predicted
            state. Unobserved velocities are initialized to 0 mean.

        """
        std_pos = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-2,
            self._std_weight_position * mean[3],
        ]
        std_vel = [
            self._std_weight_velocity * mean[3],
            self._std_weight_velocity * mean[3],
            1e-5,
            self._std_weight_velocity * mean[3],
        ]
        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        mean = np.dot(self._motion_mat, mean)
        covariance = (
            np.linalg.multi_dot((self._motion_mat, covariance, self._motion_mat.T))
            + motion_cov
        )

        return mean, covariance

    def project(self, mean, covariance):
        """Project state distribution to measurement space.

        Parameters
        ----------
        mean : ndarray
            The state's mean vector (8 dimensional array).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).

        Returns
        -------
        (ndarray, ndarray)
            Returns the projected mean and covariance matrix of the given state
            estimate.

        """
        std = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-1,
            self._std_weight_position * mean[3],
        ]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot(
            (self._update_mat, covariance, self._update_mat.T)
        )
        return mean, covariance + innovation_cov

    def update(self, mean, covariance, measurement):
        """Run Kalman filter correction step.

        Parameters
        ----------
        mean : ndarray
            The predicted state's mean vector (8 dimensional).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).
        measurement : ndarray
            The 4 dimensional measurement vector (x, y, a, h), where (x, y)
            is the center position, a the aspect ratio, and h the height of the
            bounding box.

        Returns
        -------
        (ndarray, ndarray)
            Returns the measurement-corrected state distribution.

        """
        projected_mean, projected_cov = self.project(mean, covariance)

        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False
        )
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower),
            np.dot(covariance, self._update_mat.T).T,
            check_finite=False,
        ).T
        innovation = measurement - projected_mean

        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot(
            (kalman_gain, projected_cov, kalman_gain.T)
        )
        return new_mean, new_covariance

    def gating_distance(self, mean, covariance, measurements, only_position=False):
        """Compute gating distance between state distribution and measurements.

        A suitable distance threshold can be obtained from `chi2inv95`. If
        `only_position` is False, the chi-square distribution has 4 degrees of
        freedom, otherwise 2.

        Parameters
        ----------
        mean : ndarray
            Mean vector over the state distribution (8 dimensional).
        covariance : ndarray
            Covariance of the state distribution (8x8 dimensional).
        measurements : ndarray
            An Nx4 dimensional matrix of N measurements, each in
            format (x, y, a, h) where (x, y) is the bounding box center
            position, a the aspect ratio, and h the height.
        only_position : Optional[bool]
            If True, distance computation is done with respect to the bounding
            box center position only.

        Returns
        -------
        ndarray
            Returns an array of length N, where the i-th element contains the
            squared Mahalanobis distance between (mean, covariance) and
            `measurements[i]`.

        """
        mean, covariance = self.project(mean, covariance)
        if only_position:
            mean, covariance = mean[:2], covariance[:2, :2]
            measurements = measurements[:, :2]

        cholesky_factor = np.linalg.cholesky(covariance)
        d = measurements - mean
        z = scipy.linalg.solve_triangular(
            cholesky_factor, d.T, lower=True, check_finite=False, overwrite_b=True
        )
        squared_maha = np.sum(z * z, axis=0)
        return squared_maha


def min_cost_matching(
    distance_metric,
    max_distance,
    tracks,
    detections,
    track_indices=None,
    detection_indices=None,
):
    """Solve linear assignment problem.

    Parameters
    ----------
    distance_metric : Callable[List[CustomTrack], List[Detection], List[int], List[int]) -> ndarray
        The distance metric is given a list of tracks and detections as well as
        a list of N track indices and M detection indices. The metric should
        return the NxM dimensional cost matrix, where element (i, j) is the
        association cost between the i-th track in the given track indices and
        the j-th detection in the given detection_indices.
    max_distance : float
        Gating threshold. Associations with cost larger than this value are
        disregarded.
    tracks : List[track.CustomTrack]
        A list of predicted tracks at the current time step.
    detections : List[detection.Detection]
        A list of detections at the current time step.
    track_indices : List[int]
        List of track indices that maps rows in `cost_matrix` to tracks in
        `tracks` (see description above).
    detection_indices : List[int]
        List of detection indices that maps columns in `cost_matrix` to
        detections in `detections` (see description above).

    Returns
    -------
    (List[(int, int)], List[int], List[int])
        Returns a tuple with the following three entries:
        * A list of matched track and detection indices.
        * A list of unmatched track indices.
        * A list of unmatched detection indices.

    """
    if track_indices is None:
        track_indices = np.arange(len(tracks))
    if detection_indices is None:
        detection_indices = np.arange(len(detections))

    if len(detection_indices) == 0 or len(track_indices) == 0:
        return [], track_indices, detection_indices  # Nothing to match.

    cost_matrix = distance_metric(tracks, detections, track_indices, detection_indices)
    cost_matrix[cost_matrix > max_distance] = max_distance + 1e-5

    row_indices, col_indices = linear_assignment(cost_matrix)

    matches, unmatched_tracks, unmatched_detections = [], [], []
    for col, detection_idx in enumerate(detection_indices):
        if col not in col_indices:
            unmatched_detections.append(detection_idx)
    for row, track_idx in enumerate(track_indices):
        if row not in row_indices:
            unmatched_tracks.append(track_idx)
    for row, col in zip(row_indices, col_indices):
        track_idx = track_indices[row]
        detection_idx = detection_indices[col]
        if cost_matrix[row, col] > max_distance:
            unmatched_tracks.append(track_idx)
            unmatched_detections.append(detection_idx)
        else:
            matches.append((track_idx, detection_idx))
    return matches, unmatched_tracks, unmatched_detections


def matching_cascade(
    distance_metric,
    max_distance,
    cascade_depth,
    tracks,
    detections,
    track_indices=None,
    detection_indices=None,
):
    """Run matching cascade.

    Parameters
    ----------
    distance_metric : Callable[List[CustomTrack], List[Detection], List[int], List[int]) -> ndarray
        The distance metric is given a list of tracks and detections as well as
        a list of N track indices and M detection indices. The metric should
        return the NxM dimensional cost matrix, where element (i, j) is the
        association cost between the i-th track in the given track indices and
        the j-th detection in the given detection indices.
    max_distance : float
        Gating threshold. Associations with cost larger than this value are
        disregarded.
    cascade_depth: int
        The cascade depth, should be se to the maximum track age.
    tracks : List[track.CustomTrack]
        A list of predicted tracks at the current time step.
    detections : List[detection.Detection]
        A list of detections at the current time step.
    track_indices : Optional[List[int]]
        List of track indices that maps rows in `cost_matrix` to tracks in
        `tracks` (see description above). Defaults to all tracks.
    detection_indices : Optional[List[int]]
        List of detection indices that maps columns in `cost_matrix` to
        detections in `detections` (see description above). Defaults to all
        detections.

    Returns
    -------
    (List[(int, int)], List[int], List[int])
        Returns a tuple with the following three entries:
        * A list of matched track and detection indices.
        * A list of unmatched track indices.
        * A list of unmatched detection indices.

    """
    if track_indices is None:
        track_indices = list(range(len(tracks)))
    if detection_indices is None:
        detection_indices = list(range(len(detections)))

    unmatched_detections = detection_indices
    matches = []
    for level in range(cascade_depth):
        if len(unmatched_detections) == 0:  # No detections left
            break

        track_indices_l = [
            k for k in track_indices if tracks[k].time_since_update == 1 + level
        ]
        if len(track_indices_l) == 0:  # Nothing to match at this level
            continue

        matches_l, _, unmatched_detections = min_cost_matching(
            distance_metric,
            max_distance,
            tracks,
            detections,
            track_indices_l,
            unmatched_detections,
        )
        matches += matches_l
    unmatched_tracks = list(set(track_indices) - set(k for k, _ in matches))
    return matches, unmatched_tracks, unmatched_detections


def gate_cost_matrix(
    kf,
    cost_matrix,
    tracks,
    detections,
    track_indices,
    detection_indices,
    gated_cost=INFTY_COST,
    only_position=False,
):
    """Invalidate infeasible entries in cost matrix based on the state
    distributions obtained by Kalman filtering.

    Parameters
    ----------
    kf : The Kalman filter.
    cost_matrix : ndarray
        The NxM dimensional cost matrix, where N is the number of track indices
        and M is the number of detection indices, such that entry (i, j) is the
        association cost between `tracks[track_indices[i]]` and
        `detections[detection_indices[j]]`.
    tracks : List[track.CustomTrack]
        A list of predicted tracks at the current time step.
    detections : List[detection.Detection]
        A list of detections at the current time step.
    track_indices : List[int]
        List of track indices that maps rows in `cost_matrix` to tracks in
        `tracks` (see description above).
    detection_indices : List[int]
        List of detection indices that maps columns in `cost_matrix` to
        detections in `detections` (see description above).
    gated_cost : Optional[float]
        Entries in the cost matrix corresponding to infeasible associations are
        set this value. Defaults to a very large value.
    only_position : Optional[bool]
        If True, only the x, y position of the state distribution is considered
        during gating. Defaults to False.

    Returns
    -------
    ndarray
        Returns the modified cost matrix.

    """
    gating_dim = 2 if only_position else 4
    gating_threshold = chi2inv95[gating_dim]
    measurements = np.asarray([detections[i].to_xyah() for i in detection_indices])
    for row, track_idx in enumerate(track_indices):
        track = tracks[track_idx]
        gating_distance = kf.gating_distance(
            track.mean, track.covariance, measurements, only_position
        )
        cost_matrix[row, gating_distance > gating_threshold] = gated_cost
    return cost_matrix


def _pdist(a, b):
    """Compute pair-wise squared distance between points in `a` and `b`.

    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.

    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.

    """
    a, b = np.asarray(a), np.asarray(b)
    if len(a) == 0 or len(b) == 0:
        return np.zeros((len(a), len(b)))
    a2, b2 = np.square(a).sum(axis=1), np.square(b).sum(axis=1)
    r2 = -2.0 * np.dot(a, b.T) + a2[:, None] + b2[None, :]
    r2 = np.clip(r2, 0.0, float(np.inf))
    return r2


def _cosine_distance(a, b, data_is_normalized=False):
    """Compute pair-wise cosine distance between points in `a` and `b`.

    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.
    data_is_normalized : Optional[bool]
        If True, assumes rows in a and b are unit length vectors.
        Otherwise, a and b are explicitly normalized to lenght 1.

    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.

    """
    if not data_is_normalized:
        a = np.asarray(a) / np.linalg.norm(a, axis=1, keepdims=True)
        b = np.asarray(b) / np.linalg.norm(b, axis=1, keepdims=True)
    return 1.0 - np.dot(a, b.T)


def _nn_euclidean_distance(x, y):
    """Helper function for nearest neighbor distance metric (Euclidean).

    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).

    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest Euclidean distance to a sample in `x`.

    """
    distances = _pdist(x, y)
    return np.maximum(0.0, distances.min(axis=0))


def _nn_cosine_distance(x, y):
    """Helper function for nearest neighbor distance metric (cosine).

    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).

    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest cosine distance to a sample in `x`.

    """
    distances = _cosine_distance(x, y)
    return distances.min(axis=0)


class NearestNeighborDistanceMetric(object):
    """
    A nearest neighbor distance metric that, for each target, returns
    the closest distance to any sample that has been observed so far.

    Parameters
    ----------
    metric : str
        Either "euclidean" or "cosine".
    matching_threshold: float
        The matching threshold. Samples with larger distance are considered an
        invalid match.
    budget : Optional[int]
        If not None, fix samples per class to at most this number. Removes
        the oldest samples when the budget is reached.

    Attributes
    ----------
    samples : Dict[int -> List[ndarray]]
        A dictionary that maps from target identities to the list of samples
        that have been observed so far.

    """

    def __init__(self, metric, matching_threshold, budget=None):
        if metric == "euclidean":
            self._metric = _nn_euclidean_distance
        elif metric == "cosine":
            self._metric = _nn_cosine_distance
        else:
            raise ValueError("Invalid metric; must be either 'euclidean' or 'cosine'")
        self.matching_threshold = matching_threshold
        self.budget = budget
        self.samples = {}

    def partial_fit(self, features, targets, active_targets):
        """Update the distance metric with new data.

        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : ndarray
            An integer array of associated target identities.
        active_targets : List[int]
            A list of targets that are currently present in the scene.

        """
        for feature, target in zip(features, targets):
            self.samples.setdefault(target, []).append(feature)
            if self.budget is not None:
                self.samples[target] = self.samples[target][-self.budget :]
        self.samples = {k: self.samples[k] for k in active_targets}

    def distance(self, features, targets):
        """Compute distance between features and targets.

        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : List[int]
            A list of targets to match the given `features` against.

        Returns
        -------
        ndarray
            Returns a cost matrix of shape len(targets), len(features), where
            element (i, j) contains the closest squared distance between
            `targets[i]` and `features[j]`.

        """
        cost_matrix = np.zeros((len(targets), len(features)))
        for i, target in enumerate(targets):
            cost_matrix[i, :] = self._metric(self.samples[target], features)
        return cost_matrix


def iou(bbox, candidates):
    """Computer intersection over union.

    Parameters
    ----------
    bbox : ndarray
        A bounding box in format `(top left x, top left y, width, height)`.
    candidates : ndarray
        A matrix of candidate bounding boxes (one per row) in the same format
        as `bbox`.

    Returns
    -------
    ndarray
        The intersection over union in [0, 1] between the `bbox` and each
        candidate. A higher score means a larger fraction of the `bbox` is
        occluded by the candidate.

    """
    bbox_tl, bbox_br = bbox[:2], bbox[:2] + bbox[2:]
    candidates_tl = candidates[:, :2]
    candidates_br = candidates[:, :2] + candidates[:, 2:]

    tl = np.c_[
        np.maximum(bbox_tl[0], candidates_tl[:, 0])[:, np.newaxis],
        np.maximum(bbox_tl[1], candidates_tl[:, 1])[:, np.newaxis],
    ]
    br = np.c_[
        np.minimum(bbox_br[0], candidates_br[:, 0])[:, np.newaxis],
        np.minimum(bbox_br[1], candidates_br[:, 1])[:, np.newaxis],
    ]
    wh = np.maximum(0.0, br - tl)

    area_intersection = wh.prod(axis=1)
    area_bbox = bbox[2:].prod()
    area_candidates = candidates[:, 2:].prod(axis=1)
    return area_intersection / (area_bbox + area_candidates - area_intersection)


def iou_cost(tracks, detections, track_indices=None, detection_indices=None):
    """An intersection over union distance metric.

    Parameters
    ----------
    tracks : List[deep_sort.track.CustomTrack]
        A list of tracks.
    detections : List[deep_sort.detection.Detection]
        A list of detections.
    track_indices : Optional[List[int]]
        A list of indices to tracks that should be matched. Defaults to
        all `tracks`.
    detection_indices : Optional[List[int]]
        A list of indices to detections that should be matched. Defaults
        to all `detections`.

    Returns
    -------
    ndarray
        Returns a cost matrix of shape
        len(track_indices), len(detection_indices) where entry (i, j) is
        `1 - iou(tracks[track_indices[i]], detections[detection_indices[j]])`.

    """
    if track_indices is None:
        track_indices = np.arange(len(tracks))
    if detection_indices is None:
        detection_indices = np.arange(len(detections))

    cost_matrix = np.zeros((len(track_indices), len(detection_indices)))
    for row, track_idx in enumerate(track_indices):
        if tracks[track_idx].time_since_update > 1:
            cost_matrix[row, :] = INFTY_COST
            continue

        bbox = tracks[track_idx].to_tlwh()
        candidates = np.asarray([detections[i].tlwh for i in detection_indices])
        cost_matrix[row, :] = 1.0 - iou(bbox, candidates)
    return cost_matrix


class TrackState:
    """
    Enumeration type for the single target track state. Newly created tracks are
    classified as `tentative` until enough evidence has been collected. Then,
    the track state is changed to `confirmed`. Tracks that are no longer alive
    are classified as `deleted` to mark them for removal from the set of active
    tracks.

    """

    Tentative = 1
    Confirmed = 2
    Deleted = 3


class CustomTrack:
    """
    A single target track with state space `(x, y, a, h)` and associated
    velocities, where `(x, y)` is the center of the bounding box, `a` is the
    aspect ratio and `h` is the height.

    Parameters
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.
    max_age : int
        The maximum number of consecutive misses before the track state is
        set to `Deleted`.
    feature : Optional[ndarray]
        Feature vector of the detection this track originates from. If not None,
        this feature is added to the `features` cache.
    track_state : Optional[TrackState]
        Status of the track.

    Attributes
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    hits : int
        Total number of measurement updates.
    age : int
        Total number of frames since first occurance.
    time_since_update : int
        Total number of frames since last measurement update.
    state : TrackState
        The current track state.
    features : List[ndarray]
        A cache of features. On each measurement update, the associated feature
        vector is added to this list.

    """

    def __init__(
        self,
        mean,
        covariance,
        track_id,
        n_init,
        max_age,
        feature=None,
        track_label="Object",
        track_score=1.0,
        track_state=TrackState.Tentative,
    ):
        self.mean = mean
        self.covariance = covariance
        self.track_id = track_id
        self.hits = 1
        self.age = 1
        self.time_since_update = 0
        self.track_label = track_label
        self.track_score = track_score
        self.state = track_state
        self.features = []
        if feature is not None:
            self.features.append(feature)

        self._n_init = n_init
        self._max_age = max_age

    def to_tlwh(self):
        """Get current position in bounding box format `(top left x, top left y,
        width, height)`.

        Returns
        -------
        ndarray
            The bounding box.

        """
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret

    def to_tlbr(self):
        """Get current position in bounding box format `(min x, miny, max x,
        max y)`.

        Returns
        -------
        ndarray
            The bounding box.

        """
        ret = self.to_tlwh()
        ret[2:] = ret[:2] + ret[2:]
        return ret

    def predict(self, kf):
        """Propagate the state distribution to the current time step using a
        Kalman filter prediction step.

        Parameters
        ----------
        kf : kalman_filter.KalmanFilter
            The Kalman filter.

        """
        self.mean, self.covariance = kf.predict(self.mean, self.covariance)
        self.age += 1
        self.time_since_update += 1

    def update(self, kf, detection):
        """Perform Kalman filter measurement update step and update the feature
        cache.

        Parameters
        ----------
        kf : kalman_filter.KalmanFilter
            The Kalman filter.
        detection : Detection
            The associated detection.

        """
        self.mean, self.covariance = kf.update(
            self.mean, self.covariance, detection.to_xyah()
        )
        self.features.append(detection.feature)

        self.hits += 1
        self.time_since_update = 0
        if self.state == TrackState.Tentative and self.hits >= self._n_init:
            self.state = TrackState.Confirmed

    def mark_deleted(self):
        """Mark this track as deleted."""
        self.state = TrackState.Deleted

    def mark_missed(self):
        """Mark this track as missed (no association at the current time step)."""
        if self.state == TrackState.Tentative:
            self.state = TrackState.Deleted
        elif self.time_since_update > self._max_age:
            self.state = TrackState.Deleted

    def is_tentative(self):
        """Returns True if this track is tentative (unconfirmed)."""
        return self.state == TrackState.Tentative

    def is_confirmed(self):
        """Returns True if this track is confirmed."""
        return self.state == TrackState.Confirmed

    def is_deleted(self):
        """Returns True if this track is dead and should be deleted."""
        return self.state == TrackState.Deleted


class Tracker:
    """
    This is the multi-target tracker.

    Parameters
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        A distance metric for measurement-to-track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.

    Attributes
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        The distance metric used for measurement to track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of frames that a track remains in initialization phase.
    kf : KalmanFilter
        A Kalman filter to filter target trajectories in image space.
    tracks : List[CustomTrack]
        The list of active tracks at the current time step.

    """

    def __init__(
        self, metric, max_iou_distance=0.7, max_age=70, n_init=3, update_interval=1
    ):
        self.metric = metric
        self.max_iou_distance = max_iou_distance
        self.max_age = max_age
        self.n_init = n_init

        self.kf = KalmanFilter()
        self.tracks = []
        self._next_id = 1
        self._update_interval = update_interval

    def remove(self, track_ids=[]):
        for i in range(0, len(self.tracks)):
            if self.tracks[i].id in track_ids:
                self.tracks[i].mark_deleted()

    def predict(self):
        """Propagate track state distributions one time step forward.

        This function should be called once every time step, before `update`.
        """
        for track in self.tracks:
            track.predict(self.kf)

    def update(self, detections):
        """Perform measurement update and track management.

        Parameters
        ----------
        detections : List[deep_sort.detection.Detection]
            A list of detections at the current time step.

        """
        # Run matching cascade.
        matches, unmatched_tracks, unmatched_detections = self._match(detections)

        # Update track set.
        for track_idx, detection_idx in matches:
            self.tracks[track_idx].update(self.kf, detections[detection_idx])
        for track_idx in unmatched_tracks:
            self.tracks[track_idx].mark_missed()
        for detection_idx in unmatched_detections:
            self._initiate_track(detections[detection_idx])
        self.tracks = [t for t in self.tracks if not t.is_deleted()]

        # Update distance metric.
        active_targets = [t.track_id for t in self.tracks if t.is_confirmed()]
        features, targets = [], []
        for track in self.tracks:
            if not track.is_confirmed():
                continue
            features += track.features
            targets += [track.track_id for _ in track.features]
            track.features = []
        self.metric.partial_fit(
            np.asarray(features), np.asarray(targets), active_targets
        )

    def _match(self, detections):
        def gated_metric(tracks, dets, track_indices, detection_indices):
            features = np.array([dets[i].feature for i in detection_indices])
            targets = np.array([tracks[i].track_id for i in track_indices])
            cost_matrix = self.metric.distance(features, targets)
            cost_matrix = gate_cost_matrix(
                self.kf, cost_matrix, tracks, dets, track_indices, detection_indices
            )

            return cost_matrix

        # Split track set into confirmed and unconfirmed tracks.
        confirmed_tracks = [i for i, t in enumerate(self.tracks) if t.is_confirmed()]
        unconfirmed_tracks = [
            i for i, t in enumerate(self.tracks) if not t.is_confirmed()
        ]

        # Associate confirmed tracks using appearance features.
        matches_a, unmatched_tracks_a, unmatched_detections = matching_cascade(
            gated_metric,
            self.metric.matching_threshold,
            self.max_age,
            self.tracks,
            detections,
            confirmed_tracks,
        )

        # Associate remaining tracks together with unconfirmed tracks using
        # IOU.
        iou_track_candidates = unconfirmed_tracks + [
            k for k in unmatched_tracks_a if self.tracks[k].time_since_update == 1
        ]
        unmatched_tracks_a = [
            k for k in unmatched_tracks_a if self.tracks[k].time_since_update != 1
        ]
        matches_b, unmatched_tracks_b, unmatched_detections = min_cost_matching(
            iou_cost,
            self.max_iou_distance,
            self.tracks,
            detections,
            iou_track_candidates,
            unmatched_detections,
        )

        matches = matches_a + matches_b
        unmatched_tracks = list(set(unmatched_tracks_a + unmatched_tracks_b))
        return matches, unmatched_tracks, unmatched_detections

    def _initiate_track(self, detection):
        mean, covariance = self.kf.initiate(detection.to_xyah())
        track_state = TrackState.Tentative
        if self._update_interval > 1:
            track_state = TrackState.Confirmed
        self.tracks.append(
            CustomTrack(
                mean,
                covariance,
                self._next_id,
                self.n_init,
                self.max_age,
                detection.feature,
                detection.label,
                detection.confidence,
                track_state=track_state,
            )
        )
        self._next_id += 1


def get_corrected_labels_scores(
    labels=None, scores=None, target_len=0, default_label="Object", default_score=1.0
):
    if labels is None:
        labels = [default_label] * target_len
    elif len(labels) < target_len:
        for i in range(len(labels), target_len + 1):
            labels.append(default_label)

    if scores is None:
        scores = [default_score] * target_len
    elif len(scores) < target_len:
        for i in range(len(scores), target_len + 1):
            labels.append(default_score)

    return labels, scores


class Extractor(object):
    def __init__(self, model, device):
        self.device = device
        self.net = model
        if self.net is not None:
            self.net.eval().to(self.device)
        self.size = (model.img_shape[2], model.img_shape[1])
        self.norm = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        )

    def _preprocess(self, im_crops):
        def _resize(im, size):
            return cv2.resize(im.astype(np.float32) / 255.0, size)

        im_batch = torch.cat(
            [self.norm(_resize(im, self.size)).unsqueeze(0) for im in im_crops], dim=0
        ).float()
        return im_batch

    def __call__(self, im_crops):
        im_batch = self._preprocess(im_crops)
        features = np.array([])
        with torch.no_grad():
            im_batch = im_batch.to(self.device)
            if self.net is not None:
                features = self.net(im_batch)
                features = features.cpu().numpy()
        return features


class Detection(object):
    """
    This class represents a bounding box detection in a single image.

    Parameters
    ----------
    tlwh : array_like
        Bounding box in format `(x, y, w, h)`.
    confidence : float
        Detector confidence score.
    feature : array_like
        A feature vector that describes the object contained in this image.

    Attributes
    ----------
    tlwh : ndarray
        Bounding box in format `(top left x, top left y, width, height)`.
    confidence : ndarray
        Detector confidence score.
    feature : ndarray | NoneType
        A feature vector that describes the object contained in this image.

    """

    def __init__(self, tlwh, confidence, feature, label):
        self.tlwh = np.asarray(tlwh, dtype=float)
        self.confidence = float(confidence)
        self.feature = np.asarray(feature, dtype=np.float32)
        self.label = label

    def to_tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    def to_xyah(self):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = self.tlwh.copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret


class DeepSortPredictor(object):
    def __init__(self, model, cfg, device, update_interval):
        self.min_confidence = cfg["min_confidence"]
        self.nms_max_overlap = cfg["nms_max_overlap"]
        self._update_interval = update_interval
        self.extractor = Extractor(model, device)
        max_cosine_distance = cfg["max_dist"]
        nn_budget = cfg["nn_budget"]

        metric = NearestNeighborDistanceMetric("cosine", max_cosine_distance, nn_budget)
        self.tracker = Tracker(
            metric,
            max_iou_distance=cfg["max_iou_distance"],
            max_age=cfg["max_age"],
            n_init=cfg["n_init"],
            update_interval=update_interval,
        )

    def update(self, bbox_xywh, confidences, labels, ori_img):
        outputs = []
        track_labels = []
        if len(labels) != len(confidences):
            return np.asarray(outputs), np.asarray(track_labels)

        self.height, self.width = ori_img.shape[:2]
        bbox_xywh = self.xywh_to_xywh(bbox_xywh)
        features = self._get_features(bbox_xywh, ori_img)
        bbox_tlwh = self._xywh_to_tlwh(bbox_xywh)
        detections = [
            Detection(bbox_tlwh[i], conf, features[i], labels[i])
            for i, conf in enumerate(confidences)
            if conf > self.min_confidence
        ]

        boxes = np.array([d.tlwh for d in detections])
        scores = np.array([d.confidence for d in detections])

        indices = non_max_suppression(boxes, self.nms_max_overlap, scores)
        detections = [detections[i] for i in indices]

        self.tracker.predict()
        self.tracker.update(detections)

        for track in self.tracker.tracks:
            if (
                not track.is_confirmed()
                or track.time_since_update > self._update_interval + 1
            ):
                continue
            box = track.to_tlwh()
            x1, y1, x2, y2 = self._tlwh_to_xyxy(box)
            track_id = track.track_id
            track_label = track.track_label
            track_score = track.track_score
            track_age = track.age
            track_labels.append(track_label)
            outputs.append(
                np.array(
                    [x1, y1, x2 - x1, y2 - y1, track_id, track_score, track_age],
                    dtype=float,
                )
            )
        if len(outputs) > 0:
            outputs = np.stack(outputs, axis=0)
        return outputs, np.asarray(track_labels)

    def remove(self, track_ids=[]):
        self.tracker.remove(track_ids)

    @staticmethod
    def _xywh_to_tlwh(bbox_xywh):
        if isinstance(bbox_xywh, np.ndarray):
            bbox_tlwh = bbox_xywh.copy()
        elif isinstance(bbox_xywh, torch.Tensor):
            bbox_tlwh = bbox_xywh.clone()
        bbox_tlwh[:, 0] = bbox_xywh[:, 0] - bbox_xywh[:, 2] / 2.0
        bbox_tlwh[:, 1] = bbox_xywh[:, 1] - bbox_xywh[:, 3] / 2.0
        return bbox_tlwh

    def xywh_to_xywh(self, bbox_xywh):
        bbox_xywh[:, 0] = bbox_xywh[:, 0] + bbox_xywh[:, 2] / 2.0
        bbox_xywh[:, 1] = bbox_xywh[:, 1] + bbox_xywh[:, 3] / 2.0
        return bbox_xywh

    def _xywh_to_xyxy(self, bbox_xywh):
        x, y, w, h = bbox_xywh
        x1 = max(int(x - w / 2), 0)
        x2 = min(int(x + w / 2), self.width - 1)
        y1 = max(int(y - h / 2), 0)
        y2 = min(int(y + h / 2), self.height - 1)
        return np.array([x1, y1, x2, y2])

    def _tlwh_to_xyxy(self, bbox_tlwh):
        x, y, w, h = bbox_tlwh
        x1 = max(int(x), 0)
        x2 = min(int(x + w), self.width - 1)
        y1 = max(int(y), 0)
        y2 = min(int(y + h), self.height - 1)
        return x1, y1, x2, y2

    def _xyxy_to_tlwh(self, bbox_xyxy):
        x1, y1, x2, y2 = bbox_xyxy

        t = x1
        l = y1
        w = int(x2 - x1)
        h = int(y2 - y1)
        return t, l, w, h

    def _get_features(self, bbox_xywh, ori_img):
        im_crops = []
        for box in bbox_xywh:
            x1, y1, x2, y2 = self._xywh_to_xyxy(box)
            if x2 - x1 == 0:
                x2 = x1 + 1
            if y2 - y1 == 0:
                y2 = y1 + 1
            im = ori_img[y1:y2, x1:x2, :]
            im_crops.append(im)
        if im_crops:
            features = self.extractor(im_crops)
        else:
            features = np.array([])
        return features



from __future__ import annotations
import time as _time
from typing import Union, Any
import uuid
from arcgis.gis import GIS
from .._con import Connection
from arcgis._impl.common._mixins import PropertyMap
from arcgis import env as _env
from arcgis.gis._impl._jb import StatusJob
from arcgis.gis import Item
import concurrent.futures

###########################################################################


class PortalDataStore(object):
    """

    The ``PortalDatastore`` object provides access to operations that allow you
    to manage and work with user-managed `data store items <https://enterprise.arcgis.com/en/portal/latest/use/manage-data-store-items.htm>`_ .
    Access an instance of this class using the `datastore` property of a
    :class:`~arcgis.gis.GIS` object.

    .. code-block:: python

        >>> gis = GIS("organzation url", "username", "<password>")
        >>>
        >>> portal_dstore = gis.datastore
        >>> type(portal_dstore)

        <class 'arcgis.gis._impl._datastores._ds.PortalDataStore'>

    With this class you can:

    - Validate a data store item against your server.
    - Register a data store item to your server.
    - Retrieve a list of servers your data store item is registered to.
    - Refresh your data store registration information on the server.
    - Unregister a data store from your server.
    - Publish and retrieve layers in bulk, and delete bulk-published layers.

    See `User-managed data stores <https://enterprise.arcgis.com/en/portal/latest/use/data-store-items.htm>`_
    for detailed explanation of the item types this class manages. Also,
    see `Data Item <https://developers.arcgis.com/rest/enterprise-administration/server/dataitem.htm>`_
    for technical details on managing server side components.

    .. note::
        This class provides different functionality than the :class:`~arcgis.gis.server.Datastore` object
        which is used for data stores registered directly with a :class:`~arcgis.gis.server.Server`.
    """

    _con = None
    _gis = None
    _url = None
    _properties = None
    # ----------------------------------------------------------------------

    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< PortalDataStore @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< PortalDataStore @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def describe(self, item, server_id, path, store_type="datastore"):
        """
        The ``describe`` method is used to list the contents of a data store
        added to Enterprise as a data store item. A client can use this
        method multiple times to discover the contents of the
        data store incrementally. For example, the client can request a
        description of the root, and then request sub-folders.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The data store :class:`~arcgis.gis.Item` to describe.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server the item is registered
                               to.

                               .. note::
                                   You can retrieve the `server_id` value from the dictionary
                                   returned by the federation.

                                   .. code-block:: python

                                       # Retrieve server ids
                                       >>> gis_federation = gis.admin.federation.servers
                                       >>>
                                       >>> server_list = gis_federation["servers"]
                                       >>> server_list
                                       >>>
                                       >>> [{'id': 'W2ozgK50Le9CjrI4',
                                            'name': 'mymachine.domain.com:6443',
                                            'url': 'https://mymachine.domain.com/mywebadaptor_name',
                                            'isHosted': True,
                                            'adminUrl': 'https://mymachine.domain.com:6443/arcgis',
                                            'serverRole': 'HOSTING_SERVER',
                                            'serverFunction': '',
                                            'webgisServerTrustKey': ''}]
                                       >>>
                                       >>> host_id = server_list[0]["id"]
        ------------------     --------------------------------------------------------------------
        path                   Required String. The path to any data store's root ("/"), or the
                               path to a sub-folder or entity inside the root.
        ------------------     --------------------------------------------------------------------
        store_type             Required String. For root resource the object type should be
                               `datastore`, and for sub-entities, the value depends upon data type
                               of the data store. Value can be determined by looking at
                               ``type`` values returned by :meth:`~PortalDataStore.describe`
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example: Data store item added using an enterprise geodatabase

            >>> ds_items = gis.content.search("*", item_type="Data Store")
            >>>
            >>> for ds_item in ds_items:
            >>>     print(f"{ds_item.title:33}{ds_item.id}{' '*2}{ds_item.get_data()['type']}")
            >>>
               City_Service                     ea9a628b037........7cf29eda554d3  bigDataFileShare
               CityParks2 Datastore Item API    a188c1f5b42........eb1f8dea03a8e  egdb
               various_imagery                  c6971f24a25........fe22d7142f835  folder
            >>>
            >>> egdb_dsitem = ds_items[1]
            >>>
            >>> portal_ds = gis.datastore
            >>>
            >>> describe_job = portal_ds.describe(item=egdb_dsitem,
                                                  server_id=host_id,
                                                  path="/",
                                                  store_type="datastore")
            >>>
            >>> while describe_job.status == "processing":
            >>>     continue
            >>> if describe_job.status == "succeeded":
            >>>     describe_egdb = describe_job.result()
            >>> else:
            >>>     print(f"Job status: {describe_job.status}")
            >>>     print(f"Job messages: {describe_job.messages}")
            >>>
            >>> describe_egdb

                {'id': '6bd436223f2b433285f58cf1df5bccef',
                 'definition': {'operation': 'describe',
                  'datastoreId': 'a25ae2f4c4674bf4799eb2f4fdae3b8f',
                  'serverId': 'K2ozgC10aL6ECdm2',
                  'path': '/',
                  'type': 'datastore'},
                 'status': 'succeeded',
                 'created': 166512637...1,
                 'modified': 16574263...8,
                 'messages': [],
                 'result': {'datastore': {'name': 'world_db_0klt50rsl52d4',
                   'type': 'egdb',
                   'path': '/',
                   'datastoreId': 'a25ae2f4c4674bf4799eb2f4fdae3b8f'},
                  'children': [{'name': 'world.DATAo.World_Data',
                    'displayName': 'World_Data',
                    'type': 'featureDataset',
                    'datastoreId': 'a25ae2f4c4674bf4799eb2f4fdae3b8f',
                    'path': '/world.DATAo.World_Data'},
                   {'name': 'world.DATAo.Forest_Landscapes',
                    'displayName': 'Forest_Landscapes',
                    'type': 'featureDataset',
                    'datastoreId': 'a25ae2f4c4674bf4799eb2f4fdae3b8f',
                    'path': '/world.DATAo.Forest_Landscapes'}]}}
            >>>
            >>> # describe sub-entity of the database data store
            >>>
            >>> describe_job_sub = portal_ds.describe(item=egdb_dsitem,
                                                      server_id=host_id,
                                                      path="/world.DATAo.Forest_Landscapes'",
                                                      store_type="featureDataset")

        :return:
            A :class:`~arcgis.gis._impl._jb.StatusJob` object

        """
        if isinstance(item, Item):
            item = item.id
        params = {
            "datastoreId": item,
            "serverId": server_id,
            "path": path,
            "type": store_type,
            "f": "json",
        }
        url = f"{self._url}/describe"
        res = self._con.get(url, params)
        _time.sleep(0.5)
        executor = concurrent.futures.ThreadPoolExecutor(1)
        futureobj = executor.submit(
            self._status, **{"job_id": res["jobId"], "key": res["key"]}
        )
        executor.shutdown(False)
        return StatusJob(
            future=futureobj,
            op="Describe DataStore",
            jobid=res["jobId"],
            gis=self._gis,
            notify=_env.verbose,
            extra_marker="",
            key=res.get("key", None),
        )

    # ----------------------------------------------------------------------
    def _status(self, job_id, key=None):
        """
        Checks the status of an export job

        :return: dict
        """
        params = {}
        if job_id:
            url = f"{self._gis._portal.resturl}portals/self/jobs/{job_id}"
            params["f"] = "json"
            if key:
                params["key"] = key
            res = self._con.post(url, params)
            while res["status"] not in [
                "completed",
                "complete",
                "succeeded",
            ]:
                res = self._con.post(url, params)
                if res["status"] == "failed":
                    raise Exception(res)
            count = 0
            while (
                res["status"] in ["completed", "complete", "succeeded"]
                and not "result" in res
                and count < 10
            ):
                res = self._con.post(url, params)
                if "result" in res:
                    return res
                count += 1
            return res
        else:
            raise Exception(res)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        The ``properties`` property retrieves the properties of the current
        :class:`~PortalDataStore` object

        :return:
           A dictionary of properties pertaining to the :class:`~PortalDataStore`
        """
        if self._properties is None:
            params = {"f": "json"}
            res = self._con.get(self._url, params)
            self._properties = PropertyMap(res)
        return self._properties

    # ----------------------------------------------------------------------
    def register(self, item, server_id, bind=False):
        """

        The ``register`` method allows for a data store :class:`~arcgis.gis.Item`
        that has been added to the Enterprise portal to be registered with
        Enterprise servers.

        .. note::
            Before registering a data store :class:`~arcgis.gis.Item`, it is
            recommended that you :meth:`~PortalDataStore.validate` it with
            the server.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required data store :class:`~arcgis.gis.Item` or Item Id
                               string.

                               .. note::
                                   A data store :class:`~arcgis.gis.Item` can be registered on
                                   multiple servers.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server you want to register
                               the data store item with.
        ------------------     --------------------------------------------------------------------
        bind                   Optional Boolean. Specifies whether to bind the data store item to
                               the federated server. For more information about binding a data
                               store to additional federated servers, see the note below.
                               The default value is ``False``.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example: adding data store item sourced by an enterprise
            # geodatabase and registering with server

            >>> # Get the server id value for registration
            >>> servers_dict = gis.admin.federation.servers
            >>> server_list = servers_dict["servers"]
            >>> host_id = [srvr["id"]
            >>>            for srvr in server_list
            >>>            if srvr["serverRole"] == "HOSTING_SERVER"][0]
            >>>
            >>> # Get the host server's DataStoreManager to create an
            >>> # encrypted password string for the database
            >>> host = gis.admin.servers.get(role="HOSTING_SERVER")[0]
            >>> host_dsmgr = host.datastores
            >>> conn_file_sql = r"/pathway/to/connection_file/your_connection.sde"
            >>> conn_string = host_dsmgr.generate_connection_string(conn_file_sql)
            >>>
            >>> # Add the data store item to the Enterprise Portal
            >>> text_param = {"info": {"isManaged": False,
            >>>                        "dataStoreConnectionType": "shared",
            >>>                        "connectionString": conn_string},
            >>>               "type": "egdb",
            >>>               "path": "/enterpriseDatabases/sql_server_datastore"}
            >>> item_properties = {"title": "SqlServer Datastore Item API",
            >>>                    "type": "Data Store",
            >>>                    "tags": "api_created,datastore_item,bulk_publishing",
            >>>                    "snippet": "Adding a datastore item to use api for management."}
            >>>
            >>> ds_item = gis.content.add(item_properties=item_properties,
            >>>                           text=text_param)
            >>>
            >>> # Get the Enteprises PortalDataStore and register with the server
            >>> portal_ds = gis.datastore
            >>>
            >>> portal_ds.register(item=ds_item,
            >>>                    server_id=host_id,
            >>>                    bind=False)

        :return:
            A boolean indicating success (True), or failure (False)

        .. note::
            To create a data store :class:`~arcgis.gis.Item` from a data store previously registered
            with an Enterprise Server, see the `Create a data store item from an existing data store <https://enterprise.arcgis.com/en/portal/latest/administer/windows/create-item-from-existing-data-store.htm#ESRI_SECTION1_58D081604CF841AC80D527D34A67660C>`_ page in the Enterprise Portal documentation.

        """
        if isinstance(item, Item):
            item_id = item.id
        elif isinstance(item, str):
            item_id = item
        url = "{base}/addToServer".format(base=self._url)
        params = {"f": "json", "datastoreId": item_id, "serverId": server_id}
        if not bind is None:
            params["bindToServerDatastore"] = bind
        res = self._con.post(url, params)
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    @property
    def _all_datasets(self):
        """
        The _all_datasets resource page provides access to bulk publishing
        operations. These operations allow users to publish and synchronize
        datasets from a given data store, return a list of published layers,
        and remove all previously published layers in preparation for the
        deregistration of a data store.

        :return: Boolean

        """
        url = "{base}/allDatasets".format(base=self._url)
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "success" in res:
            return res["success"]
        elif "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def delete_layers(self, item):
        """
        The ``delete_layers`` method removes all layers published from a database
        data store :class:`~arcgis.gis.Item`.

        .. note::
            Before a data store :class:`~arcgis.gis.Item` can be unregistered
            from a server, all of its bulk-published layers must be deleted.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The database data store
                               :class:`~arcgis.gis.Item` from which to delete all
                               the published layers.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example

            >>> # Get data store item sourced by a database
            >>> db_dsitem = gis.content.get("<item_id of data store item>")
            >>>
            >>> portal_ds = gis.datastore
            >>> portal_ds.delete_layers(db_dsitem)

        :return:
            A boolean indicating success (``True``), or failure (``False``)


        """
        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)
        params = {"f": "json", "datastoreId": item_id}
        url = f"{self._url}/allDatasets/deleteLayers"
        res = self._con.post(url, params)
        if res["success"] == True:
            status = item.status()
            while status["status"].lower() != "completed":
                status = item.status()
                if status["status"].lower() == "failed":
                    return False
                else:
                    _time.sleep(2)
            return True
        return False

    # ----------------------------------------------------------------------
    def layers(self, item):
        """
        The ``layers`` operation returns a list of layers bulk published from
        a database data store item using the
        :meth:`~PortalDataStore.publish_layers` method.

        .. note::
            The ``layers`` method returns a list of dictionaries. Each
            dictionary contains a `layer` key and a `dataset` key for each
            layer created during publishing.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The data store :class:`~arcgis.gis.Item` to list all
                               published layers and registered datasets.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example
            >>> ds_items = gis.content.get("*", item_type="Data Store")
            >>>
            >>> db_dsitem = [ds
            >>>              for ds in ds_items
            >>>              if ds.get_data()["type"] == "egdb"][0]
            >>>
            >>> portal_ds = gis.datastore
            >>>
            >>> portal_ds.layers(db_dsitem)

            [{'layer': {'id': '059...711',
                        'title': 'Boundary_GTM',
                        'created': 1667433642607,
                        'modified': 1667433732631,
                        'url': 'https://<url>/<webadaptor>/rest/services/<folder_name>/Boundary_GTM/FeatureServer',
                        'type': 'Feature Service'},
              'dataset': {'name': 'world.dto.CountryBoundary_GTM',
                          'displayName': 'CountryBoundary_Guatemala',
                          'type': 'featureClass',
                          'datastoreId': 'ad38...01b8e',
                          'path': '/world.dto.GTM_Datasets/world.dto.Boundary_GTM'}},
                          .
                          .
                          .
             {'layer': {'id': '0d5a6e6f2003483ba987eb044ca4dec6',
                        'title': 'dk_lau_2016',
                        'created': 1667433476227,
                        'modified': 1667433605914,
                        'url': 'https://<url>/<webadaptor>/rest/services/<folder_name>/dk_lau_2016/MapServer',
                        'type': 'Map Service'},
              'dataset': {'name': 'world.dto.dk_lau_2016',
                          'displayName': 'dk_lau_2016',
                          'type': 'table',
                          'datastoreId': 'ad38...01b8e',
                          'path': '/world.dto.dk_lau_2016'}}]

        :return:
            A list of dictionaries. For each layer, a dictionary of information about the
            layer and the source dataset from which it was published.
        """

        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)

        url = "{base}/allDatasets/getLayers".format(base=self._url)
        params = {"f": "json", "datastoreId": item_id}
        res = self._con.post(url, params)
        if "layerAndDatasets" in res:
            return res["layerAndDatasets"]
        return []

    # ----------------------------------------------------------------------
    def publish(
        self,
        config: dict,
        server_id,
        folder=None,
        description=None,
        tags: Union[list, str] = None,
    ):
        """
        The ``publish`` operation is used to publish services by reference from
        specific datasets in a data store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        config                 Required Dictionary.  This is the service configuration property
                               and it must contain the reference to the data in the data store. It
                               specifies the data store Id and the path of the data.  A client can
                               discover the proper paths of the data by using the
                               :meth:`~PortalDataStore.describe` method.

                               .. note::
                                   The ``cacheStoreId`` value for this dictionary correpsonds to the
                                   datastore id value.

                               .. code-block:: python

                                   # Example format

                                   >>> config = {"type":"SceneServer",
                                   >>>           "serviceName":"sonoma",
                                   >>>           "properties":{"pathInCachedStore":"/v17_i3s/SONOMA_LiDAR.i3srest",
                                   >>>                         "cacheStoreId":"d7b072...00d9"}}
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique Id of the server to publish to.

                               .. note::
                                   Any :class:`~arcgis.gis.server.Server` id can be obtained from
                                   the dictionary returned by by the
                                   :attr:`~arcgis.gis.admin.Federation.servers` property of
                                   :class:`~arcgis.gis.admin.Federation` objects.
        ------------------     --------------------------------------------------------------------
        folder                 Optional String. The name of the folder on the server to store the
                               service. If none is provided, it is placed in the root.
        ------------------     --------------------------------------------------------------------
        description            Optional String. An optional string to attach to the
                               generated :class:`~arcgis.gis.Item`.
        ------------------     --------------------------------------------------------------------
        tags                   Optional list. An array of descriptive words that describes the
                               newly published :class:`~arcgis.gis.Item`.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example

            >>> portal_ds = gis.datastore
            >>>
            >>> service_config = {"type":"SceneServer",
            >>>                   "serviceName":"sonoma",
            >>>                   "properties":{"pathInCachedStore":"/v17_i3s/SONOMA_LiDAR.i3srest",
            >>>                                 "cacheStoreId":"d7b072...00d9"}}
            >>>
            >>> server_list = gis.admin.Federation.servers["servers"]
            >>> gis_server_id = [s["id"]
            >>>                  for s in server_list
            >>>                  if s["serverRole"] == "HOSTING_SERVER"]
            >>>
            >>> pub_job = portal_ds.publish(config= service_config,
            >>>                             server_id= gis_server_id)
            >>>
            >>> if pub_job.status == "succeeded":
            >>>     published_item = pub_job.result()

        :return:
            A :class:`~arcgis.gis._impl._jb.StatusJob` object

        """
        url = f"{self._url}/publish"
        if isinstance(tags, list):
            tags = ",".join([str(t) for t in tags])

        params = {
            "serviceConfiguration": config,
            "serverId": server_id,
            "serverFolder": folder,
            "description": description,
            "tags": tags,
            "f": "json",
        }
        res = self._con.post(url, params)
        executor = concurrent.futures.ThreadPoolExecutor(1)
        futureobj = executor.submit(
            self._status, **{"job_id": res["jobId"], "key": res["key"]}
        )
        executor.shutdown(False)
        return StatusJob(
            future=futureobj,
            op="Publish",
            jobid=res["jobId"],
            gis=self._gis,
            notify=_env.verbose,
            extra_marker="",
            key=res.get("key", None),
        )

    # ----------------------------------------------------------------------
    def servers(self, item):
        """
        The ``servers`` property returns a list of your servers that a given
        data store has been registered to. This operation returns the
        serverId, the server name, both the server and admin URLs, and
        whether or not the server is hosted.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The data store :class:`~arcgis.gis.Item` to
                               for which to list all registered servers.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example

            >>> # Get a data store item
            >>> ds_item = gis.content.search("*", item_type="Data Store)[0]
            >>>
            >>> server_list = gis.datastore.servers(ds_item)
            >>> server_list[0]

            [{'id': 'T4omqC59LaH4vLi0',
              'name': 'jserver.domain.com:6443',
              'adminUrl': 'https://server.domain.com:6443/arcgis',
              'url': 'https://jserver.domain.com/server',
              'isHosted': True,
              'serverType': 'ArcGIS',
              'serverRole': 'HOSTING_SERVER',
              'serverFunction': ''}]

        :return:
            A list of dictionaries with metadata about each server the item is
            registered with.

        """

        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)

        url = self._url + "/getServers"
        params = {"f": "json", "datastoreId": item_id}
        res = self._con.post(url, params)
        if "servers" in res:
            return res["servers"]
        return res

    # ----------------------------------------------------------------------
    def publish_layers(
        self,
        item: Item,
        srv_config: dict[str, Any],
        server_id: str,
        *,
        folder: str | None = None,
        server_folder: str | None = None,
        sync_metadata: bool | None = None,
        use_config: bool | None = None,
    ):
        """
        The ``publish_layers`` operation publishes, or syncs, the datasets from a
        database data store :class:`~arcgis.gis.Item`. This results in at least
        one layer per dataset.

        .. note::
            When this operation is called for the first time, an **argument for
            every parameter must be provided**. On subsequent calls, this method
            will synchronize the datasets in the data store with the layers
            creted in the Enterprise, which includes both publishing layers
            from newly added datasets and removing layers for datasets no
            longer found in the data store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The data store :class:`~arcgis.gis.Item` holding
                               the content to publish.
        ------------------     --------------------------------------------------------------------
        srv_config             Required Dict. The JSON that will be used as a template for all the
                               services that will be published or synced. This JSON can be used to
                               change the properties of the output map and feature services.

                               .. note::
                                   Only map service configurations with feature services enabled are
                                   supported by this parameter.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The serverId that the datasets will be published to.

                               .. note::
                                   Use `gis.admin.federation.servers` to get the id values for
                                   individual servers federated with the Enterprise portal.
        ------------------     --------------------------------------------------------------------
        folder                 Required String. The folder to which the datasets will be published.

                               .. note::
                                   This folder must exist in the Enteprise portal.
        ------------------     --------------------------------------------------------------------
        server_folder          Required String. The name of the server folder.

                               .. note::
                                   If this folder does not exist, the method will
                                   create it.
        ------------------     --------------------------------------------------------------------
        sync_metadata          Optional bool. Determines if item info details are updated using the
                               metadata of the source dataset when a sync is performed. The default
                               is false.
        ------------------     --------------------------------------------------------------------
        use_config             Optional bool. When true, the new `srv_config` will be applied to
                               all layers.
        ==================     ====================================================================

        :return:
            Boolean object.

        .. code-block:: python

            # Usage Example: bulk publishing from enterprise geodatabase data store item

            >>> ds_items = gis.content.search("*", item_type="Data Store")
            >>>
            >>> db_dsitem = [ds for ds in ds_items if ds.get_data()["type"] == "egdb"][0]
            >>>
            >>> portal_folderid = [f["id"]
            >>>                    for f in gis.users.me.folders
            >>>                    if f["title"] == "My_Bulk_Layers_Folder"]
            >>>
            >>> service_template = {"serviceName": None,
            >>>                     "type": "MapServer",
            >>>                     "capabilities":"Map,Query",
            >>>                     "extensions": [{"typeName": "FeatureServer",
            >>>                                     "capabilities":"Query,Create,Update,Delete",
            >>>                                     "enabled": "true",
            >>>                                     "properties": {"maxRecordCount": 3500}}]}
            >>>
            >>>
            >>> portal_ds = gis.datastore
            >>>
            >>> bulk_publish_job = portal_ds.publish_layers(item = db_dsitem,
            >>>                                             srv_config = service_template,
            >>>                                             server_id = host_id,
            >>>                                             folder = portal_folderid,
            >>>                                             server_folder="bulk_egdb_layers")
            >>> bulk_publish_job
            True

        """
        if server_folder is None:
            base = "buld_pub_"
            if isinstance(item, Item):
                base = item.title.lower().replace(" ", "")
            server_folder = f"{base}{uuid.uuid4().hex[:3]}"
        if folder is None:
            from arcgis.gis import UserManager, User, ContentManager

            isinstance(self._gis, GIS)
            cm = self._gis.content
            folder = cm.create_folder(folder=f"srvc_folder_{uuid.uuid4().hex[:5]}")[
                "id"
            ]
        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)
        url = self._url + "/allDatasets/publishLayers"
        params = {
            "f": "json",
            "datastoreId": item_id,
            "templateSvcConfig": srv_config,
            "portalFolderId": folder,
            "serverId": server_id,
            "serverFolder": server_folder,
        }
        if not sync_metadata is None and isinstance(sync_metadata, bool):
            params["syncItemInfo"] = sync_metadata
        if not use_config is None and isinstance(use_config, bool):
            params["applySvcConfigChanges"] = use_config
        res = self._con.post(url, params)
        if res["success"] == True:
            status = item.status()
            while status["status"].lower() != "completed":
                status = item.status()
                if status["status"].lower() == "failed":
                    return False
                else:
                    _time.sleep(2)
            return True
        return False

    # ----------------------------------------------------------------------
    def unregister(self, item, server_id):
        """
        The ``unregister`` method removes the data store from the list of data
        stores on the server.

        .. code-block:: python

            # Usage Example
            >>> ds_item = gis.content.get("<item id of data store>")
            >>>
            >>> server_info = gis.admin.federation.servers["servers"][0]
            >>> server_id = server_info["id"]
            >>>
            >>> gis.datastore.unregister(item = dsitem,
            >>>                          server_id = server_id)

        :return:
            A boolean indicating success (True), or failure (False)

        """
        if isinstance(item, Item):
            item_id = item.id
        elif isinstance(item, str):
            item_id = item
        url = self._url + "/removeFromServer"
        params = {"f": "json", "datastoreId": item_id, "serverId": server_id}
        res = self._con.post(url, params)
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    def refresh_server(self, item, server_id):
        """
        The ``refresh_server`` method updates the server with information that
        changed in the data store. See
        `Manage data store items <https://enterprise.arcgis.com/en/portal/latest/use/manage-data-store-items.htm>`_
        for more information.

        .. note::
            After a data store :class:`~arcgis.gis.Item` has been registered, there
            may be times in which the registration information may be changed. When
            changes like these occur, the server will need to be updated with
            the newly configured information so that your users will still be
            able to access the data store item without interruption.

        .. warning::
            This operation can only be performed after the data
            store information has been updated.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required data store :class:`~arcgis.gis.Item` or item id (as string).
                               The data store to register with the server. Note that a data store
                               can be registered on multiple servers.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server you want to register
                               the datastore with.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example
            >>> portal_ds = gis.datastore
            >>>
            >>> ds_item = gis.content.search("*", item_type="Data Store")[0]
            >>> server_list = gis.admin.federation.servers["servers]
            >>>
            >>> host_id = [s["id"]
            >>>            for s in server_list
            >>>            if s["serverRole"] == "HOSTING_SERVER][0]
            >>>
            >>> portal_ds.refresh_server(item = ds_item, server_id = host_id)

        :return:
            A boolean indicating success (True), or failure (False)

        """
        if isinstance(item, Item):
            item_id = item.id
        elif isinstance(item, str):
            item_id = item

        url = self._url + "/refreshServer"
        params = {"f": "json", "datastoreId": item_id, "serverId": server_id}
        res = self._con.post(url, params)
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    def validate(self, server_id, item=None, config=None, future=False):
        """
        The ``validate`` method ensures that your ArcGIS Server can connect and use
        use the datasets accessed by a given data store item. The data store
        :class:`~arcgis.gis.Item` can be validated by using either the `id`
        property, or an item object tself as the `item` argument.

        .. note::
            While this operation can be called before or after the data store item
            has been registered, it is recommended to validate before
            registration on the server.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server with which you want to
                               register the data store.
        ------------------     --------------------------------------------------------------------
        item                   Optional. The item id or data store
                               :class:`~arcgis.gis.Item` to validate. Required if no ``config``
                               provided.

                               .. note::
                                   A single data store item can be registered on multiple servers.
        ------------------     --------------------------------------------------------------------
        config                 Optional dict. The connection information for a new datastore.
                               Required if no ``item`` provided.
        ------------------     --------------------------------------------------------------------
        future                 Optional bool. Indicates whether to run the validate operation
                               asynchronously. The default is `False`.
        ==================     ====================================================================

        .. code-block:: python

            # Usage Example: Validating an added item against the Enterprise Raster Analytics server

            >>> gis = GIS(url="<url to GIS>", username="<username>", password="<password>")
            >>>
            >>> ds_item = gis.content.search("*", item_type="data store")[0]
            >>>
            >>> server_list = gis.admin.federation.servers["servers]
            >>> raster_id = [srv["id"]
            >>>              for srv in server_list
            >>>              if srv["function"] == "RasterAnalytics"][0]
            >>>
            >>> portal_ds = gis.datastore
            >>> portal_ds.validate(server_id = raster_id,
            >>>                    item = ds_item)

        :return:
            A boolean indicating success (True), or failure (False)

        """

        if item and isinstance(item, Item):
            item_id = item.id
        elif item and isinstance(item, str):
            item_id = item
        else:
            item_id = None

        url = self._url + "/validate"
        params = {"f": "json", "serverId": server_id}
        if item:
            params["datastoreId"] = item_id
        elif config:
            import json

            params["datastore"] = json.dumps(config)
        else:
            raise ValueError("Invalid parameters, an item or config is required.")
        res = self._con.post(url, params)
        if "status" in res:
            if res["status"] == "success":
                return True
            return res["status"]
        return res



import os, copy
from typing import Dict, List, Any, Optional, Union, Iterator
import pandas as pd
from arcgis.gis import GIS
from arcgis.geometry import Geometry
from arcgis import env as _env
from arcgis._impl.common._isd import InsensitiveDict
from functools import lru_cache


###########################################################################
class OGCCollection:
    """
    Represents a single OGC dataset

    ================  ===============================================================================
    **Parameter**      **Description**
    ----------------  -------------------------------------------------------------------------------
    url               Required String. The web address endpoint.
    ----------------  -------------------------------------------------------------------------------
    gis               Optional :class:`~arcgis.gis.GIS`. The connection object.
    ================  ===============================================================================

    """

    _gis = None
    _url = None
    _properties = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS = None) -> "OGCCollection":
        """Constructor"""
        assert (
            str(url).lower().find("ogcfeatureserver") > -1
            and os.path.basename(url).isdigit()
        )
        if gis is None:
            gis = _env.active_gis or GIS()
        self._gis = gis
        self._url = url

    # ---------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """returns the service properties"""
        params = {"f": "json"}
        try:
            res = self._gis._con.get(self._url, params)
            return InsensitiveDict(res)
        except:
            res = self._gis._con.post(self._url, params)
            return InsensitiveDict(res)

    # ---------------------------------------------------------------------
    def _process_row(self, feature: Dict[str, Any]) -> Dict[str, Any]:
        """Converts the GeoJSON Geometry to Esri JSON and format the row accordingly"""
        row = {"SHAPE": Geometry(feature["geometry"]) if feature["geometry"] else None}
        row.update(feature["properties"])
        return row

    # ---------------------------------------------------------------------
    def query(
        self,
        query: Optional[str] = None,
        limit: int = 10000,
        bbox: Optional[List[float]] = None,
        bbox_sr: Optional[int] = None,
        time_filter: Optional[str] = None,
        return_all: bool = False,
        **kwargs,
    ) -> Union[Dict[str, Any], pd.DataFrame]:
        """
        Queries the :class:`~arcgis.mapping.ogc.OGCFeatureService` Layer and returns back the information as a Spatially Enabled DataFrame.

        ================  ===============================================================================
        **Parameter**      **Description**
        ----------------  -------------------------------------------------------------------------------
        query             Optional String. A SQL based query applied to the service.
        ----------------  -------------------------------------------------------------------------------
        limit             Optional Integer. The number of records to limit to.  The default is 10,000.
        ----------------  -------------------------------------------------------------------------------
        bbox              Optional List[float]. The bounding box to limit search in.
        ----------------  -------------------------------------------------------------------------------
        bbox_sr           Optional Integer. The coordinate reference system as a WKID.
        ----------------  -------------------------------------------------------------------------------
        time_filter       Optional String. The dates to filter time by.
        ================  ===============================================================================

        :return: Union[Dict[str, Any], pd.DataFrame]
        """
        url = f"{self._url}/items"
        params = {"f": "json"}
        if query:
            params["filter"] = query
        if isinstance(limit, int) and limit >= 0:
            params["limit"] = limit
        if bbox and bbox_sr:
            params["bbox"] = bbox
            params["bbox-crs"] = bbox_sr
        if kwargs.get("crs", None):
            params["crs"] = kwargs.pop("crs")
        params["offset"] = kwargs.pop("offset", 0)
        if time_filter:
            params["datetime"] = time_filter
        for k, v in kwargs.items():
            params[k] = v
        as_dict = kwargs.pop("as_dict", False)
        if as_dict == False:  # returns all records as sedf
            res = self._gis._con.get(url, params)
            if len(res["features"]) == 0:
                return pd.DataFrame([])
            results = [
                self._process_row(feature) for feature in res["features"] if feature
            ]
            params["offset"] += limit
            while True:
                res = self._gis._con.get(url, params)
                if res == {} or res["numberReturned"] == 0:
                    break
                elif return_all == False and len(results) >= limit:
                    results = results[:limit]
                    break
                elif res["numberReturned"] < limit:
                    r1 = [
                        self._process_row(feature)
                        for feature in res["features"]
                        if feature
                    ]
                    results.extend(r1)
                    break
                else:
                    r1 = [
                        self._process_row(feature)
                        for feature in res["features"]
                        if feature
                    ]
                    results.extend(r1)
                params["offset"] += limit
            df = pd.DataFrame(results)
            df.spatial.name
            return df
        else:
            results = self._gis._con.get(url, params)
            if len(results["features"]) == 0:
                return results
            res = copy.deepcopy(results)
            params["offset"] += limit
            while res["numberReturned"] > 0:
                res = self._gis._con.get(url, params)

                if res == {} or res["numberReturned"] == 0:
                    break

                results["features"].extend(res["features"])
                if return_all == False and len(results["features"]) >= limit:
                    results["features"] = results["features"][:limit]
                    break
                elif res["numberReturned"] < limit:
                    break
                params["offset"] += limit
            return results
        return {}

    # ---------------------------------------------------------------------
    def get(self, feature_id: int) -> Dict[str, Any]:
        """
        Gets an individual feature on the service. Needs to correspond
        to an id of the feature.

        :return: Dict[str, Any]
        """
        assert isinstance(feature_id, int)
        url = f"{self._url}/items/{feature_id}"
        params = {"f": "json"}
        return self._gis._con.get(url, params)


###########################################################################
class OGCFeatureService:
    """
    Represents the Hosted OGC Feature Server

    ================  ===============================================================================
    **Parameter**      **Description**
    ----------------  -------------------------------------------------------------------------------
    url               Required String. The web address endpoint.
    ----------------  -------------------------------------------------------------------------------
    gis               Optional :class:`~arcgis.gis.GIS`. The connection object.
    ================  ===============================================================================

    """

    _gis = None
    _url = None
    _properties = None

    # ---------------------------------------------------------------------
    def __init__(self, url, gis=None) -> "OGCFeatureService":
        """Constructor"""
        assert str(url).lower().endswith("ogcfeatureserver")
        if gis is None:
            gis = _env.active_gis or GIS()
        self._gis = gis
        self._url = url

    # ---------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """returns the service properties"""
        params = {"f": "json"}
        if self._properties is None:
            try:
                res = self._gis._con.get(self._url, params)
                self._properties = InsensitiveDict(res)
            except:
                res = self._gis._con.post(self._url, params)
                self._properties = InsensitiveDict(res)
        return self._properties

    # ---------------------------------------------------------------------
    @property
    @lru_cache(maxsize=100)
    def conformance(self) -> Dict[str, Any]:
        """
        Provides the API conformance with the OGC standard.

        :return: Dict[str, Any]
        """
        url = f"{self._url}/conformance"
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    # ---------------------------------------------------------------------
    @property
    def collections(self) -> Iterator[OGCCollection]:
        """
        Yields all the OGC Feature Service Layers within the service.

        :return: Iterator[:class:`~arcgis.mapping.ogc.OGCCollection`]
        """
        url = f"{self._url}/collections"
        params = {"f": "json"}
        for idx, lyr in enumerate(self._gis._con.get(url, params)["collections"]):
            service_url = f"{url}/{lyr['id']}"
            yield OGCCollection(url=service_url, gis=self._gis)



# necessary import

import os
from pathlib import Path
import traceback
import warnings
import glob

try:
    import numpy as np
    import torch
    from .._utils.common import ArcGISMSImage
    from fastai.vision.data import ImageList, ItemBase
    from ..models._maskrcnn_utils import ArcGISImageSegment as ImageMask
    from ..models._unet_utils import ArcGISImageSegment as ImageSegment
    from ..models._unet_utils import is_contiguous, map_to_contiguous

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class MaskLabelSemanticItem(ItemBase):
    "ItemBase class suitable for Panoptic(Mask + Semantic)"

    def __init__(self, mask, labels, semantic):
        self.obj = (mask, labels, semantic)
        self.data = [mask.data, labels, semantic.data]
        self.shape = semantic.shape

    def apply_tfms(self, tfms, **kwargs):
        self.obj = [
            self.obj[0].apply_tfms(tfms, **kwargs),
            self.obj[1],
            self.obj[2].apply_tfms(tfms, **kwargs),
        ]
        self.data = [self.obj[0].data, self.obj[1], self.obj[2].data]
        return self

    def __repr__(self):
        return f"{self.__class__.__name__}{(self.obj[0].shape, self.obj[1].shape, self.obj[2].shape)}"


# Class to read Panoptic Segmentation Labels
class PanopticSegmentationLabelList(ImageList):
    "ItemList class suitable for Panoptic Targets (Labels)."

    def __init__(
        self,
        items,
        chip_size,
        classes=None,
        class_mapping=None,
        color_mapping=None,
        **kwargs,
    ):
        # max number of masks (K in ground truth, N  in predictions)
        self.K = kwargs.pop("n_masks")
        self.inst_class_mapping = kwargs.pop("inst_class_mapping")
        self.inv_inst_class_mapping = {v: k for k, v in self.inst_class_mapping.items()}

        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append("classes")
        self.classes = classes
        self.chip_size = chip_size

        # self.items[i] - WindowsPath('data/panoptic_kent_100/labels/000000102.tif')
        fn, ext = os.path.splitext(os.path.split(self.items[0])[1])
        self.inst_lbl_path = self.items[0].parent.parent / "labels2"

        # Create a list of instance class values
        self.instance_classes = list(self.inst_class_mapping.keys())

        ## Check whether the class values are contiguous and create contiguous mapping
        self.is_contiguous = is_contiguous(
            sorted([0] + list(self.class_mapping.keys()))
        )
        if not self.is_contiguous:
            self.pixel_mapping = [0] + list(self.class_mapping.keys())
            self.indexed_inst_classes = map_to_contiguous(
                torch.tensor(self.instance_classes), self.pixel_mapping
            ).tolist()
        else:
            self.indexed_inst_classes = self.instance_classes

        # Create a list of files in all class folders in labels2
        self.inst_lbl_files = list()
        for root, dirs, files in os.walk(self.inst_lbl_path):
            for file in files:
                if file.endswith(ext):
                    self.inst_lbl_files.append(os.path.join(root, file))

    def get(self, i):
        # Use the default method to open semantic labels
        semantics = super().get(i)
        # Create individual masks of all semantic segments
        masks, labels = self.create_semantic_masks(semantics.data)

        # self.items[i] - WindowsPath('data/panoptic_kent_100/labels/000000102.tif')
        fn, ext = os.path.splitext(os.path.split(self.items[i])[1])

        # Create a list of instance masks for the current image
        mask_files = list()
        for file in self.inst_lbl_files:
            if fn in file:
                mask_files.append(file)

        if len(mask_files) != 0:
            inst_masks, inst_labels = self.create_instance_masks(mask_files)
            masks = torch.cat((masks, inst_masks))
            labels = torch.cat((labels, inst_labels))

        n_labels = labels.shape[0]
        if n_labels > self.K:
            warnings.warn(
                "{} classes and instance labels present in one of the chips but only {} masks created. Please rerun 'prepare_data' with a higher 'n_masks' value.".format(
                    n_labels, self.K
                ),
                stacklevel=1,
            )
            masks = masks[: self.K]
            labels = labels[: self.K]

        # Add padding to masks and labels
        if n_labels < self.K:
            pad_size = self.K - n_labels
            mask_pad = torch.zeros(
                (pad_size, masks.shape[1], masks.shape[2]), dtype=torch.uint8
            )
            masks = torch.cat((masks, mask_pad), dim=0)
            labels_pad = torch.zeros(pad_size, dtype=torch.long)
            labels = torch.cat((labels, labels_pad), dim=0)

        masks = ArcGISMSImage(masks.to(torch.float))
        return MaskLabelSemanticItem(masks, labels, semantics)

    def open(self, fn):
        x = ArcGISMSImage.open(fn).data
        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)
        return ImageSegment(x, color_mapping=self.color_mapping)

    # Method to convert the semantics to individual masks and labels
    def create_semantic_masks(self, semantic):
        # Change the instance class pixels to -1
        for inst_cls in self.indexed_inst_classes:
            semantic = torch.where(
                semantic == int(inst_cls), torch.tensor(-1), semantic
            )

        semantic_np = np.asarray(semantic.cpu())
        labels = np.unique(semantic.cpu())

        ## TODO: investigate this
        # Remove the label -1 (instance classes)
        if labels[0] == -1:
            labels = labels[1:]

        semantic_labels = torch.tensor(labels, dtype=torch.long)

        if len(labels) == 0:
            return (
                torch.empty(size=(0, semantic.shape[1], semantic.shape[2])),
                semantic_labels,
            )

        semantic_masks = semantic_np == labels[:, None, None]
        semantic_masks = [mask.astype(np.uint8) for mask in semantic_masks]
        semantic_masks = torch.from_numpy(np.stack(semantic_masks, axis=0))
        return semantic_masks, semantic_labels

    # Method to convert the instance label files to masks and labels
    def create_instance_masks(self, fn):
        masks = []
        labels = []

        for mask_file in fn:
            # Read the mask file and grab the classname of the mask
            mask_img = torch.from_numpy(
                ArcGISMSImage.read_image(Path(mask_file)).astype("int16")
            )
            lbl_name = Path(mask_file).parent.name
            lbl_value = self.inv_inst_class_mapping[lbl_name]

            # Ensure a channel dimension exists
            if (
                len(mask_img.shape) == 3
            ):  # if mask shape has a channels dim (i.e. #channels > 1)
                mask_img = mask_img.permute(2, 0, 1)
            else:
                mask_img = torch.unsqueeze(mask_img, 0)

            for ch in range(mask_img.shape[0]):
                unique_instances = np.unique(mask_img[ch]).max()

                # For each unique mask id, starting 1, create an individual mask
                for instance in range(1, unique_instances + 1):
                    instance_mask = (mask_img[ch] == instance).to(torch.uint8)
                    masks.append(instance_mask)
                    labels.append(lbl_value)

        masks = torch.stack(masks)
        labels = torch.tensor(labels)

        if not self.is_contiguous:
            labels = map_to_contiguous(labels, self.pixel_mapping)

        return masks, labels

    def analyze_pred(self, pred, thresh=0.5, model=None):
        pred = model._model_conf.post_process(pred, thresh)
        return pred


class PanopticSegmentationItemList(ImageList):
    "`ItemList` suitable for Panoptic tasks."
    _label_cls = PanopticSegmentationLabelList
    _square_show_res = False
    _div = None
    _imagery_type = None

    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)


def compute_n_masks(path):
    n_masks = 0

    with open(path / "map.txt") as f:
        line = f.readline()
        ext = line.split()[1].split(".")[-1].lower()

    imgs = glob.glob(str(path / "images") + "/*.{}".format(ext))
    inst_labels = glob.glob(str(path / "labels2") + "/*/*.{}".format(ext))

    # Read labels and labels2 for each image
    for i, img in enumerate(imgs):
        label_count = 0
        img_name, extn = os.path.splitext(os.path.split(img)[1])
        sem = path / "labels" / "{}.{}".format(img_name, ext)
        insts = [i for i in inst_labels if img_name in i]

        # Count number of classes in semantics
        if not os.path.exists(sem):
            continue
        semantic = ArcGISMSImage.open(sem).data
        label_count = len(np.unique(semantic.cpu()))

        # Count number of instances
        for inst in insts:
            mask_img = torch.from_numpy(
                ArcGISMSImage.read_image(Path(inst)).astype("int16")
            )
            # Ensure a channel dimension exists
            if len(mask_img.shape) == 3:
                # if mask shape has a channels dim (i.e. #channels > 1)
                mask_img = mask_img.permute(2, 0, 1)
            else:
                mask_img = torch.unsqueeze(mask_img, 0)

            for ch in range(mask_img.shape[0]):
                unique_instances = np.unique(mask_img[ch]).max().item()
                label_count += unique_instances

        n_masks = max(n_masks, label_count)

    # Return max n_mask
    return n_masks



from arcgis.apps.dashboard.dashboard import Dashboard
from arcgis.apps.dashboard._serialchart._serial_chart import SerialChart
from arcgis.apps.dashboard._header._header import Header
from arcgis.apps.dashboard._sidepanel.side_panel import SidePanel
from arcgis.apps.dashboard._side_widgets._addon import (
    CategorySelector,
    NumberSelector,
    DatePicker,
)
from arcgis.apps.dashboard._embcontent._embcontent import EmbeddedContent
from arcgis.apps.dashboard._richtext._richtext import RichText
from arcgis.apps.dashboard._details._details import Details
from arcgis.apps.dashboard._indicator._indicator import (
    Indicator,
    IndicatorData,
    ReferenceData,
)
from arcgis.apps.dashboard._piechart._pie_chart import PieChart
from arcgis.apps.dashboard._map_legend._map_legend import MapLegend
from arcgis.apps.dashboard._gauge._gauge import Gauge
from arcgis.apps.dashboard._list._list import List
from arcgis.apps.dashboard._utils._basewidget import add_row, add_column



from ._codetemplate import image_translation_prf
import json
import traceback
from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel, _EmptyData

try:
    from ._pix2pix_utils import (
        pix2pixLoss,
        Pix2PixPerceptualLoss,
        pix2pixTrainer,
        Pix2PixPerceptualTrainer,
        optim,
        compute_metrics,
        compute_fid_metric,
    )
    from ._pix2pix_utils import pix2pix as pix2pix_model
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from .._data_utils.pix2pix_data import show_results, predict
    from pathlib import Path
    from fastai.vision import DatasetType, Learner, partial
    import torch

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class Pix2Pix(ArcGISModel):

    """
    Creates a model object which generates fake images of type B from type A.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch with image chip sizes
                            in multiples of 256. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the :class:`~arcgis.learn.Pix2Pix`, which
                            is UNet with vanilla encoder by default.
                            Compatible backbones as encoder: 'resnet18', 'resnet34',
                            'resnet50', "resnet101", "resnet152", 'resnext50', 'wide_resnet50'
    ---------------------   -------------------------------------------
    perceptual_loss         Optional boolean. True when Perceptual loss is used.
                            Default set to False.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.Pix2Pix` Object
    """

    def __init__(
        self,
        data,
        pretrained_path=None,
        backbone=None,
        perceptual_loss=False,
        *args,
        **kwargs
    ):
        super().__init__(data, backbone, **kwargs)
        self._check_dataset_support(data)
        if self._data.chip_size % 256 == 0:
            bnds = ["o" for i in range(self._data.n_channel)]
            if not hasattr(self._data, "_bands"):
                self._data._bands = bnds
            else:
                if not self._data._bands:
                    self._data._bands = bnds
            self._data._extract_bands = list(range(self._data.n_channel))
            pix2pix_gan = pix2pix_model(
                self._data,
                self._data.n_channel,
                self._data.n_channel,
                self._backbone if backbone else None,
                perceptual_loss,
                self._data.chip_size,
            )
            if perceptual_loss:
                self.learn = Learner(
                    data,
                    pix2pix_gan,
                    loss_func=Pix2PixPerceptualLoss(pix2pix_gan),
                    opt_func=partial(optim.Adam, betas=(0.5, 0.99)),
                    callback_fns=[Pix2PixPerceptualTrainer],
                )
            else:
                self.learn = Learner(
                    data,
                    pix2pix_gan,
                    loss_func=pix2pixLoss(pix2pix_gan),
                    opt_func=partial(optim.Adam, betas=(0.5, 0.99)),
                    callback_fns=[pix2pixTrainer],
                )

            self.learn.model = self.learn.model.to(self._device)
            self._slice_lr = False
            self.perceptual_loss = perceptual_loss
            self.backbone = backbone.lower() if backbone else backbone
            if pretrained_path is not None:
                self.load(pretrained_path)
            self._code = image_translation_prf

            def __str__(self):
                return self.__repr__()

            def __repr__(self):
                return "<%s>" % (type(self).__name__)

        else:
            raise Exception("Image chip sizes should be in multiples of 256")

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.Pix2Pix` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.Pix2Pix` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        resize_to = emd.get("resize_to")
        chip_size = emd["ImageHeight"]

        if data is None:
            if emd.get("IsMultispectral", False):
                data = _EmptyData(
                    path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
                )
                data = get_multispectral_data_params_from_emd(data, emd)
                data._is_multispectral = emd.get("IsMultispectral", False)
                normalization_stats_b = dict(emd.get("NormalizationStats_b"))
                for _stat in normalization_stats_b:
                    if normalization_stats_b[_stat] is not None:
                        normalization_stats_b[_stat] = torch.tensor(
                            normalization_stats_b[_stat]
                        )
                    setattr(data, ("_" + _stat), normalization_stats_b[_stat])

            else:
                data = _EmptyData(
                    path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
                )

            data.n_channel = emd.get("n_intput_channel", None)
            if data.n_channel == None:
                data.n_channel = emd.get("n_channel", None)
            data.emd_path = emd_path
            data.emd = emd
            data._is_empty = True
            data.resize_to = chip_size

        model_params["backbone"] = emd.get("backbone", None)
        model_params["perceptual_loss"] = emd.get("perceptual_loss", False)
        data._extract_bands = emd.get("extract_bands", None)
        data._bands = emd.get("bands", None)
        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        return self.compute_metrics(show_progress=True)

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_pix2pix"
        _emd_template["perceptual_loss"] = self.perceptual_loss
        _emd_template["backbone"] = self.backbone
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageTranslation.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageTranslation.py"
        _emd_template["ModelType"] = "Pix2Pix"
        _emd_template["n_intput_channel"] = self._data.n_channel
        _emd_template["NormalizationStats_b"] = {
            "band_min_values": self._data._band_min_values_b,
            "band_max_values": self._data._band_max_values_b,
            "band_mean_values": self._data._band_mean_values_b,
            "band_std_values": self._data._band_std_values_b,
            "scaled_min_values": self._data._scaled_min_values_b,
            "scaled_max_values": self._data._scaled_max_values_b,
            "scaled_mean_values": self._data._scaled_mean_values_b,
            "scaled_std_values": self._data._scaled_std_values_b,
        }
        for _stat in _emd_template["NormalizationStats_b"]:
            if _emd_template["NormalizationStats_b"][_stat] is not None:
                _emd_template["NormalizationStats_b"][_stat] = _emd_template[
                    "NormalizationStats_b"
                ][_stat].tolist()
        _emd_template["n_channel"] = len(
            _emd_template["NormalizationStats_b"]["band_min_values"]
        )
        _emd_template["extract_bands"] = self._data._extract_bands
        _emd_template["bands"] = self._data._bands
        return _emd_template

    def show_results(self, rows=2, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================


        **kwargs**

        =====================   ===========================================
        rgb_bands               Optional list of integers (band numbers)
                                to be considered for rgb visualization.
        =====================   ===========================================

        """
        show_results(self, rows, **kwargs)

    def predict(self, path):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        =====================   ===========================================

        """
        return predict(self, path)

    def compute_metrics(self, show_progress=True):
        """
        Computes Peak Signal-to-Noise Ratio (PSNR) and
        Structural Similarity Index Measure (SSIM) on validation set.
        Additionally, computes Frechet Inception Distance (FID) for
        RGB imagery only.

        """
        psnr, ssim = compute_metrics(self, self._data.valid_dl, show_progress)
        if self._data._imagery_type_b == "RGB" and self._data.n_channel == 3:
            fid = compute_fid_metric(self, self._data)
            return {
                "PSNR": "{0:1.4e}".format(psnr),
                "SSIM": "{0:1.4e}".format(ssim),
                "FID": "{0:1.4e}".format(fid),
            }
        else:
            fid = None
            return {"PSNR": "{0:1.4e}".format(psnr), "SSIM": "{0:1.4e}".format(ssim)}

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return Pix2Pix._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Pix2Pix", "Export_Tiles"]

    @property
    def supported_backbones(self):
        """
        Supported backbones for this model.
        """
        return Pix2Pix._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [
            "resnet18",
            "resnet34",
            "resnet50",
            "resnet101",
            "resnet152",
            "resnext50",
            "wide_resnet50",
        ]



import os
import sys
import json
import uuid
from arcgis.gis import GIS
from arcgis.mapping import create_symbol
from arcgis._impl.common._isd import InsensitiveDict
from ._base import BaseOGC


###########################################################################
class GeoRSSLayer(BaseOGC):
    """
    The GeoRSSLayer class is used to create a layer based on GeoRSS. GeoRSS is a
    way to add geographic information to an RSS feed. The GeoRSSLayer supports
    both GeoRSS-Simple and GeoRSS GML encodings, and multiple geometry types.

    It exports custom RSS tags as additional attribute fields in the form of
    simple strings or an array of JSON objects.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required String. The URL of the GeoRSS sevice.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    line_symbol         Optional Dict. The symbol for the polyline data in the GeoRSS.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    point_symbol        Optional Dict. The symbol for the point data in the GeoRSS.
    ---------------     --------------------------------------------------------------------
    polygon_symbol      Optional Dict. The symbol for the polygon data in the GeoRSS.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and LayerList widgets.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ===============     ====================================================================

    """

    _line_symbol = None
    _point_symbol = None
    _polygon_symbol = None
    _type = "GeoRSS"

    # ----------------------------------------------------------------------
    def __init__(self, url, **kwargs):
        super(GeoRSSLayer, self)
        self._id = kwargs.pop("id", uuid.uuid4().hex)
        self._url = url
        scale = kwargs.pop("scale", (0, 0))
        assert isinstance(scale, (list, tuple)) and len(scale) == 2
        self._min_scale = scale[0]
        self._max_scale = scale[1]
        self._opacity = kwargs.pop("opacity", 1)
        self._copyright = kwargs.pop("copyright", None)
        self._title = kwargs.pop("title", None)
        self.point_symbol = kwargs.pop("point_symbol", None)
        self.polygon_symbol = kwargs.pop("polygon_symbol", None)
        self.line_symbol = kwargs.pop("line_symbol", None)
        self.title = kwargs.pop("title", "GeoRSS Feed")

    # ----------------------------------------------------------------------
    @property
    def point_symbol(self) -> InsensitiveDict:
        """
        Gets/Sets the Point Symbol for Point Geometries

        :return:
            ``InsensitiveDict``: A case-insensitive ``dict`` like object used to update and alter JSON
            A variants of a case-less dictionary that allows for dot and bracket notation.
        """
        if self._point_symbol is None:
            self._point_symbol = InsensitiveDict(
                dict(create_symbol(geometry_type="point"))
            )
        return self._point_symbol

    # ----------------------------------------------------------------------
    @point_symbol.setter
    def point_symbol(self, value: dict):
        """
        Gets/Sets the Point Symbol for Point Geometries

        :return:
            ``InsensitiveDict``: A case-insensitive ``dict`` like object used to update and alter JSON
            A variants of a case-less dictionary that allows for dot and bracket notation.
        """
        if isinstance(value, dict):
            self._point_symbol = InsensitiveDict(value)
        elif value is None:
            self._point_symbol = InsensitiveDict(
                dict(create_symbol(geometry_type="point"))
            )
        else:
            self._point_symbol = InsensitiveDict(value)

    # ----------------------------------------------------------------------
    @property
    def line_symbol(self) -> InsensitiveDict:
        """
        Gets/Sets the Line Symbol for Polyline Geometries

        :return:
            ``InsensitiveDict``: A case-insensitive ``dict`` like object used to update and alter JSON
            A variants of a case-less dictionary that allows for dot and bracket notation.
        """
        if self._line_symbol is None:
            self._line_symbol = InsensitiveDict(
                dict(create_symbol(geometry_type="polyline"))
            )
        return self._line_symbol

    # ----------------------------------------------------------------------
    @line_symbol.setter
    def line_symbol(self, value: dict):
        """
        Gets/Sets the Line Symbol for Polyline Geometries

        :return:
            ``InsensitiveDict``: A case-insensitive ``dict`` like object used to update and alter JSON
            A variants of a case-less dictionary that allows for dot and bracket notation.
        """
        if isinstance(value, dict):
            self._line_symbol = InsensitiveDict(value)
        elif value is None:
            self._line_symbol = InsensitiveDict(
                dict(create_symbol(geometry_type="polyline"))
            )
        else:
            self._line_symbol = InsensitiveDict(value)

    # ----------------------------------------------------------------------
    @property
    def polygon_symbol(self) -> InsensitiveDict:
        """
        Gets/Sets the Polygon Symbol for Polygon Geometries

        :return:
            ``InsensitiveDict``: A case-insensitive ``dict`` like object used to update and alter JSON
            A variants of a case-less dictionary that allows for dot and bracket notation.
        """
        if self._polygon_symbol is None:
            self._polygon_symbol = InsensitiveDict(
                dict(create_symbol(geometry_type="polygon"))
            )
        return self._polygon_symbol

    # ----------------------------------------------------------------------
    @polygon_symbol.setter
    def polygon_symbol(self, value: dict):
        """
        Gets/Sets the Polygon Symbol for Polygon Geometries

        :return:
            ``InsensitiveDict``: A case-insensitive ``dict`` like object used to update and alter JSON
            A variants of a case-less dictionary that allows for dot and bracket notation.
        """
        if isinstance(value, dict):
            self._polygon_symbol = InsensitiveDict(value)
        elif value is None:
            self._polygon_symbol = InsensitiveDict(
                dict(create_symbol(geometry_type="polygon"))
            )
        else:
            self._polygon_symbol = InsensitiveDict(value)

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self) -> dict:
        """Represents the MapView's widget JSON format"""
        add_layer = {
            "type": self._type,
            "url": self._url,
            "opacity": self.opacity,
            "minScale": self.scale[0],
            "maxScale": self.scale[1],
            "pointSymbol": self.point_symbol._json(),
            "polygonSymbol": self.polygon_symbol._json(),
            "lineSymbol": self.line_symbol._json(),
            "id": self._id,
            "title": self.title,
        }
        return add_layer

    @property
    def _operational_layer_json(self) -> dict:
        """Represents the WebMap's JSON format"""
        return self._lyr_json
